#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import Generator, os

# class Qt
class Qt:
	# ctor
	def __init__( self, target, makefile ):
		self.makefile = makefile

		path = self.getQtPath()
		target.include( os.path.join( self.makefile.get( 'Qt' ), 'include' ) )

	# getQtPath
	def getQtPath( self ):
		path = self.makefile.get( 'Qt' )
		assert path, 'no Qt path set'

		path = os.path.expanduser( path )
		assert os.path.exists( path ), 'Qt path {0} doesnt exist'.format( path )

		return path

	# framework
	def framework( self, name ):
		if not name.startswith( 'Qt' ):
			return name

		return os.path.join( self.getQtPath(), 'lib', name + '.framework' )

	# wrap
	def wrap( self, target, generator, fileName, extension ):
		if not os.path.exists( fileName ):
			return False

		if extension == '.ui':
			self.uic( target, generator, fileName )
			return True
		elif extension == '.qrc':
			self.rcc( target, generator, fileName )
			return True
		elif extension == '.h':
			self.moc( target, generator, fileName )
			return True

		return False

	# moc
	def moc( self, target, generator, fileName ):
		fp = open( fileName )

		if fp.read().find( 'Q_OBJECT' ) != -1:
			name, ext 	 	= os.path.splitext( os.path.basename( fileName ) )
			outputFileName 	= os.path.join( generator.getPathForTarget( target ), 'moc' + name )
			cmd 			= '{0}/bin/moc -o {1}.cpp {2}'.format( self.makefile.get( 'Qt' ), outputFileName, fileName )

			target.command( [ fileName ], outputFileName, cmd, 'Generating moc{0}.cpp...'.format( name ), ['.cpp'] )

		fp.close()

	# uic
	def uic( self, target, generator, fileName ):
		name, ext 	 	= os.path.splitext( os.path.basename( fileName ) )
		outputFileName 	= os.path.join( generator.getPathForTarget( target ), 'ui' + name )
		cmd 			= '{0}/bin/uic -o {1}.h {2}'.format( self.makefile.get( 'Qt' ), outputFileName, fileName )

		target.command( [ fileName ], outputFileName, cmd, 'Generating ui{0}.cpp...'.format( name ), ['.h'] )

	# rcc
	def rcc( self, target, generator, fileName ):
		name, ext 	 	= os.path.splitext( os.path.basename( fileName ) )
		outputFileName 	= os.path.join( generator.getPathForTarget( target ), 'rc' + name )
		cmd 		 	= '{0}/bin/rcc -name {1} -o {2}.cpp {3}'.format( self.makefile.get( 'Qt' ), name, outputFileName, fileName )

		# Extract file names
		fp    = open( fileName )
		data  = fp.read()
		start = 0
		input = []
		while True:
			start = data.find( '<file>', start )
			if start == -1:
				break

			end = data.find( '</file>', start )
			input.append( os.path.join( os.path.dirname( fileName ), data[start + 6:end] ) )
			start = end
		fp.close()

		target.command( [ fileName ] + input, outputFileName, cmd, 'Generating rc{0}.cpp...'.format( name ), ['.cpp'] )


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

from Make import Make

# class HTML5
class HTML5( Make ):
	# constructor
	def __init__( self ):
		Make.__init__( self )

		self.extensions = { 'shared': 'so', 'static': 'a', 'executable': 'html' }
		self.toolchain  = { 'cc': 'emcc', 'cxx': 'em++', 'ar': 'llvm-ar' }
		
	# generate
	def generate( self ):
		print( 'Generating HTML5 project...' )
		Make.generate( self )


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os, shutil

from Library    import LocalLibrary
from Library    import ExternalLibrary
from Framework  import Framework
from Folder     import Folder
from ..Makefile import Makefile

# class Target
class Target:
	Ident           = 0
	StaticLibrary   = 'static'
	SharedLibrary   = 'shared'
	Executable      = 'executable'

	# ctor
	def __init__( self, name, sources = None, include = None, defines = None, linkTo = None, libs = None ):
		self.message( 'Configure ' + name + '...' )

		self.project        = Makefile.getProject()
		self.name           = name
		self.defines        = []
		self._root          = Folder( self )
		self.includes       = []
		self.resources      = []
		self.commands       = []
		self._libraries     = []
		self._frameworks    = []
		self.type           = linkTo

		self._currentSourceDir = Makefile.getCurrentSourceDir()
		self._currentBinaryDir = Makefile.getCurrentBinaryDir()

		# Register this target
		if self.project:
			self.project.registerTarget( self )

		# Add default include folders
		if include:
			for path in include:
				self.include( path )

		# Add default libs
		if libs:
			for lib in libs:
				if isinstance( lib, Framework ):
					self.frameworks( lib )
				if isinstance( lib, Target ):
					self.link( lib.name )
				else:
					self.link( lib )

		# Add default defines
		if defines:
			for define in defines:
				self.define( define )

		# Add default sources
		if sources:
			for source in sources:
				fullPath = self.toFullPath( source )

				if os.path.isfile( fullPath ):
					self.files( source )
				else:
					self.dirs( source )

		Target.Ident = Target.Ident + 1

		# Link to
		if linkTo == Target.Executable:
			self.executable()
		elif linkTo == Target.StaticLibrary:
			self.staticLibrary()
		elif linkTo == Target.SharedLibrary:
			self.sharedLibrary()

	# binaryPath
	@property
	def binaryPath( self ):
		return self.project.generator.getPathForTarget( self ) if self.project else self._currentBinaryDir

	# sourcePath
	@property
	def sourcePath( self ):
		return self._currentSourceDir

	# toFullPath
	def toFullPath( self, path ):
		return os.path.join( self.sourcePath, Makefile.substituteVars( path ) )

	# toSourcePath
	def toSourcePath( self, path ):
		return os.path.relpath( path, self.sourcePath )

	# define
	def define( self, *list ):
		[self.defines.append( define ) for define in list]

	# dirs
	def dirs( self, *list ):
		[self._root.addFilesFromDirectory( self.toFullPath( path ) ) for path in list]

	# files
	def files( self, *list ):
		[self._root.addFileAtPath( self.toSourcePath( self.toFullPath( path ) ) ) for path in list]

	# link
	def link( self, *list ):
		[self._libraries.append( item if isinstance( item, ExternalLibrary ) else LocalLibrary( self, item ) ) for item in list]

	# frameworks
	def frameworks( self, *list ):
		[self._frameworks.append( Framework( self, item ) if isinstance( item, str ) else item ) for item in list]

	# include
	def include( self, *list ):
		[self.includes.append( self.toFullPath( path ) ) for path in list]

	# assets
	def assets( self, *list ):
		[self.resources.append( path ) for path in list]

	# filterSourceFiles
	def filterSourceFiles( self, filter = None ):
		return self._root.filterFiles( filter )

	# filterFrameworks
	def filterFrameworks( self, filter = None ):
		return [framework for framework in self._frameworks if filter == None or filter( framework )]

	# filterLibraries
	def filterLibraries( self, filter = None ):
		return [library for library in self._libraries if filter == None or filter( library )]

	# sharedLibrary
	def sharedLibrary( self ):
		self.message( 'Configured as shared library' )
		self.type = Target.SharedLibrary

		Target.Ident = Target.Ident - 1

	# staticLibrary
	def staticLibrary( self ):
		self.message( 'Configured as static library' )
		self.type = Target.StaticLibrary

		Target.Ident = Target.Ident - 1

	# executable
	def executable( self, **params ):
		self.message( 'Configured as executable' )
		self.type   = Target.Executable
		self.params = params

		Target.Ident = Target.Ident - 1

	# message
	@classmethod
	def message( self, text ):
		msg = ''
		for i in range( 0, Target.Ident * 4 ):
			msg += ' '
		print( msg + text )

'''
import string, os, glob, Generator, Extensions

# class Target
class Target:
	Ident = 0

	Extensions = {
		'RTTI': Extensions.RTTI,
	    'Qt':   Extensions.Qt
	}

	# ctor
	def __init__( self, name, project = None ):
		self.message( 'Configure ' + name + '...' )

		self.project 	= project
		self.name		= name
		self.type		= None
		self.params		= {}
		self.ext        = []
		self.sourcePath	= Generator.getCurrentSourceDir()

		if project:
			self.binaryPath = project.generator.getPathForTarget( self )
			self.ext.extend( project.ext )
		else:
			self.binaryPath = Generator.getCurrentBinaryDir()

		self.sources 	    = []
		self.javaPackages   = []
		self.commands 	    = []
		self.libs 		    = []
		self.defines   	    = []
		self.includes 	    = []
		self.groups		    = None
		self.resources      = None

		self.include( self.binaryPath )

		# WORKAROUND
		if project:
			project.include( project.generator.getPathForTarget( self ) )

		# Register this target
		if project:
			project.registerTarget( self )

		Target.Ident = Target.Ident + 1

	# getGenerator
	def getGenerator( self ):
		return self.project.generator if self.project else self.generator

	# registerExtension
	def registerExtension( self, name, ext ):
		Target.Extensions[name] = ext

	# extensions
	def extensions( self, *list ):
		for ext in list:
			if not ext in Target.Extensions.keys():
				raise Exception( 'Unknown extension {0}'.format( ext ) )

			self.message( 'Extension {0} enabled'.format( ext ) )
			self.ext.append( Target.Extensions[ext]( self, Generator ) )

	# message
	@classmethod
	def message( self, text ):
		msg = ''
		for i in range( 0, Target.Ident * 4 ):
			msg += ' '
		print( msg + text )

	# group
	def group( self, path, files ):
		# class Group
		class Group:
			# ctor
			def __init__( self, target, name, parent = None ):
				self.target     = target
				self.name 		= name
				self.parent  	= parent
				self.path 		= name if parent == None else os.path.join( parent.path, name )
				self.groups		= {}
				self.files 		= []

		if self.groups == None:
			self.groups = Group( self, self.name )

		# getOrCreateGroup
		def getOrCreateGroup( path ):
			group = self.groups

			for name in path.replace( '\\', '/' ).split( '/' ):
				if not name in group.groups.keys():
					group.groups[name] = Group( self, name, group )

				group = group.groups[name]

			return group

		if path == '.':
			path = 'Code'
		elif path != 'GeneratedFiles':
			path = 'Code/' + path

		insertTo = getOrCreateGroup( path )

		for file in files:
			fileName = os.path.relpath( file, self.getGenerator().getPathForTarget( self ) )
			if not fileName in insertTo.files:
				insertTo.files.append( fileName )
			else:
				print 'WARNING: file {0} duplicate found inside group {1}'.format( os.path.basename( fileName ), path )

	# embed
	def embed( self, fileName, identifier ):
		fileName 		= os.path.join( self.sourcePath, fileName )
		name, ext 	 	= os.path.splitext( os.path.basename( fileName ) )
		outputFileName	= os.path.join( self.getGenerator().getPathForTarget( self ), 'embedded' + name.title() )
		cmd 		 	= 'python ' + os.path.join( Generator.SourceDir, '../tools/scripts/EmbedFile.py' ) + ' ' + fileName + ' ' + outputFileName + '.h' + ' ' + 'embedded' + name.title()

		self.command( [ fileName ], outputFileName, cmd, '(Embedding)...', ['.h'] )

	# java
	def java( self, package ):
		if not package.startswith( self.sourcePath ):
			package = os.path.join( self.sourcePath, package )

		self.javaPackages.append( package )

	# files
	def files( self, *list ):
		# Workaround for self.dirs method
		if type( list[0] ) == tuple:
			list = list[0]
		#################################

		for inputFileName in list:
			fileName  = inputFileName.replace( '\\', '/' )
			extension = os.path.splitext( fileName )[1]

			if not fileName.startswith( self.sourcePath ):
				fileName = os.path.join( self.sourcePath, fileName )

			# Check extensions
			wrapped = False
			for ext in self.ext:
				wrapped = wrapped or ext.wrap( self, self.getGenerator(), fileName, extension )

			if not wrapped and not extension in ['.c', '.cpp', '.h', '.m', '.mm', '.cxx', '.inl', '.plist']:
				continue

			if fileName.find( self.binaryPath ) == -1:
				folder = os.path.relpath( os.path.dirname( fileName ), self.sourcePath )
				self.group( folder, [ fileName ] )
				
			self.sources.append( fileName )

	# command
	def command( self, input, output, command, message, generatedExtensions = None ):
		# class Command
		class Command:
			def __init__( self ):
				self.output  				= output
				self.message 				= message
				self.command 				= command
				self.input	 				= input
				self.generatedExtensions	= generatedExtensions

		self.commands.append( Command() )

		if generatedExtensions != None:
			for ext in generatedExtensions:
				self.sources.append( os.path.join( self.binaryPath, os.path.basename( output ) ) + ext )
				self.group( 'GeneratedFiles', [ output + ext ] )

	# dirs
	def dirs( self, *list ):
		for inputFolder in list:
			path = os.path.join( self.sourcePath, inputFolder )

			# Asterisk folder path
			if path.endswith( '/*' ):
				for folder in glob.glob( path ):
					if os.path.isdir( folder ):
						self.dirs( os.path.relpath( folder, Generator.getCurrentSourceDir() ) )

				continue

			# Get the file list
			fileList = tuple( glob.glob( path + '/*.*' ) )

			# Add files to project
			self.files( fileList )

	# link
	def link( self, *list ):
		for lib in list:
			self.message( 'Dependent on ' + lib )
			self.libs.append( { 'name': lib, 'framework': False } )

	# frameworks
	def frameworks( self, *list ):
		for framework in list:
			# Check extensions
			for ext in self.ext:
				if hasattr( ext, 'framework' ):
					framework = ext.framework( framework )


			self.message( 'Uses ' + framework )
			self.libs.append( { 'name': framework, 'framework': True } )

	# define
	def define( self, *list ):
		for define in list:
			self.defines.append( define )

	# include
	def include( self, *list ):
		for path in list:
			path = string.replace( path, '\\', '/' )
			if not os.path.isabs( path ):
				self.includes.append( os.path.join( Generator.getCurrentSourceDir(), path ) )
			else:
				self.includes.append( path )

	# assets
	def assets( self, path ):
		self.message( 'Assets ' + path )
		self.resources = path

	# sharedLibrary
	def sharedLibrary( self ):
		self.message( 'Configured as shared library' )
		self.type = 'shared'

		Target.Ident = Target.Ident - 1

	# staticLibrary
	def staticLibrary( self ):
		self.message( 'Configured as static library' )
		self.type = 'static'

		Target.Ident = Target.Ident - 1

	# executable
	def executable( self, **params ):
		self.message( 'Configured as executable' )
		self.type   = 'executable'
		self.params = params

		Target.Ident = Target.Ident - 1
'''


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os

import BuildPhases, Files, Targets

from ...Template        import Template
from ObjectList         import ObjectList
from Group              import Group
from BuildConfiguration import BuildConfiguration
from ConfigurationList  import ConfigurationList

# class Objects
class Objects:
	# ctor
	def __init__( self, project ):
		self.groups             = ObjectList( 'PBXGroup' )
		self.configurations     = ObjectList( 'PBXConfiguration' )
		self.configurationLists = ObjectList( 'PBXConfigurationList' )
		self.buildPhases        = ObjectList( 'PBXBuildPhase' )
		self.targets            = ObjectList( 'PBXNativeTarget' )
		self.buildFiles         = ObjectList( 'PBXBuildFiles' )
		self.fileReferences     = ObjectList( 'PBXFileReference' )
		self.proxies            = ObjectList( 'PBXContainerItemProxy' )
		self.dependencies       = ObjectList( 'PBXTargetDependency' )
		self.project            = project

	# createGroup
	def createGroup( self, name = None, parent = None ):
		group = Group( self, name )

		if parent:
			parent.children.add( group )

		self.groups.add( group )
		return group

	# createFramework
	def createFramework( self, name ):
		if name.endswith( '.framework' ):
			return self.createFileReference( 'wrapper.framework', name, '"<group>"' )
		else:
			return self.createFileReference( 'wrapper.framework', 'System/Library/Frameworks/{0}.framework'.format( name ), 'SDKROOT' )

	# createBuildFile
	def createBuildFile( self, target, file ):
		result = Files.BuildFile( target, file )
		self.buildFiles.add( result )
		return result

	# createFileReference
	def createFileReference( self, type, path, dir ):
		result = Files.FileReference( type, path, dir )
		self.fileReferences.add( result )
		return result

	# createSourceBuildPhase
	def createSourceBuildPhase( self, target ):
		result = BuildPhases.SourceBuildPhase( target )
		self.buildPhases.add( result )
		return result

	# createResourcePhase
	def createResourcePhase( self, target ):
		result = BuildPhases.ResourceBuildPhase( target )
		self.buildPhases.add( result )
		return result

	# createShellScriptPhase
	def createShellScriptPhase( self, target, command ):
		result = BuildPhases.ShellScriptPhase( target, command )
		self.buildPhases.add( result )
		return result

	# createFrameworkPhase
	def createFrameworkPhase( self, target ):
		result = BuildPhases.FrameworkBuildPhase( target )
		self.buildPhases.add( result )
		return result

	# createDependency
	def createDependency( self, target ):
		result = Targets.Dependency( self, target )
		self.dependencies.add( result )
		return result

	# createContainerItemProxy
	def createContainerItemProxy( self, target ):
		result = Targets.ContainerItemProxy( target, self.project )
		self.proxies.add( result )
		return result

	# createConfiguration
	def createConfiguration( self, name, settings ):
		configuration = BuildConfiguration( name, settings )
		self.configurations.add( configuration )

		return configuration

	# createStaticLibrary
	def createStaticLibrary( self, group, name, path, configurations, defaultConfiguration ):
		result = Targets.StaticLibrary( self, group, name, path, configurations, defaultConfiguration )
		self.targets.add( result )

		return result

	# createApplication
	def createApplication( self, group, name, path, configurations, defaultConfiguration ):
		result = Targets.Application( self, group, name, path, configurations, defaultConfiguration )
		self.targets.add( result )

		return result

	# createApplicationBundle
	def createApplicationBundle( self, group, name, path, configurations, defaultConfiguration, teamIdentifier ):
		result = Targets.ApplicationBundle( self, group, name, path, configurations, defaultConfiguration, teamIdentifier )
		self.targets.add( result )

		return result

	# createConfigurationList
	def createConfigurationList( self, target, configurations, default ):
		result = ConfigurationList( default, target )

		for k, v in configurations.items():
			result.add( self.createConfiguration( k, v ) )

		self.configurationLists.add( result )

		return result

	# compile
	def compile( self ):
		items  = [self.buildFiles, self.proxies, self.fileReferences, self.groups, self.targets, self.project, self.buildPhases, self.dependencies, self.configurations, self.configurationLists ]
		result = ''

		for objects in items:
			result += Template( Objects.Section ).compile( { 'name': objects.isa, 'items': objects.compile() } )

		return result

	# Section
	Section = """
/* Begin {name} section */
{items}
/* End {name} section */
"""


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os

from Resource       import Resource
from Objects        import Objects
from ObjectList     import ObjectList
from ...Template    import Template

# class Project
class Project( Resource ):
	# ctor
	def __init__( self, name, settings ):
		Resource.__init__( self, 'PBXProject', name )

		self.organization       = 'YAP'
		self.archiveVersion     = 1
		self.objectVersion      = 46
		self.objects            = Objects( self )
		self.targets            = ObjectList()
		self.configurationList  = self.objects.createConfigurationList( self, settings, 'Debug' )
		self.mainGroup          = self.objects.createGroup()
		self.sourcesGroup       = self.addGroup( 'Sources' )
		self.frameworksGroup    = self.addGroup( 'Frameworks' )
		self.productsGroup      = self.addGroup( 'Products' )

	# addGroup
	def addGroup( self, name ):
		return self.objects.createGroup( name, self.mainGroup )

	# addStaticLibrary
	def addStaticLibrary( self, name, path, settings, defaultConfiguration ):
		target = self.objects.createStaticLibrary( self.sourcesGroup, name, path, settings, defaultConfiguration )
		self.targets.add( target )
		self.productsGroup.add( target.product )

		return target

	# addApplication
	def addApplication( self, name, path, settings, defaultConfiguration ):
		target = self.objects.createApplication( self.sourcesGroup, name, path, settings, defaultConfiguration )
		self.targets.add( target )
		self.productsGroup.add( target.product )

		return target

	# addApplicationBundle
	def addApplicationBundle( self, name, path, settings, defaultConfiguration, teamIdentifier ):
		target = self.objects.createApplicationBundle( self.sourcesGroup, name, path, settings, defaultConfiguration, teamIdentifier )
		self.targets.add( target )
		self.productsGroup.add( target.product )

		return target

	# addFramework
	def addFramework( self, name ):
		framework = self.objects.createFramework( name )
		self.frameworksGroup.add( framework )
		return framework

	# addProjectLibrary
	def addProjectLibrary( self, name ):
		library = self.objects.createFileReference( 'archive.ar', 'lib' + name + '.a', 'BUILT_PRODUCTS_DIR' )
		self.frameworksGroup.add( library )
		return library

	# addLibrary
	def addLibrary( self, name ):
		library = self.objects.createFileReference( 'archive.ar', name, '"<group>"' )
		self.frameworksGroup.add( library )
		return library

	# compileTargetAttributes
	def compileTargetAttributes( self ):
		result = ''
		for target in self.targets.items:
			result += target.compileAttributes()
		return result

	# generate
	def generate( self ):
		return Template( Project.Root ).compile( { 'id': self.id, 'archive.version': self.archiveVersion, 'object.version': self.objectVersion, 'objects': self.objects.compile() } )

	# compile
	def compile( self ):
		return Template( Project.Project ).compile( {
														'id':                   self.id,
														'name':                 self.name,
		                                                'isa':                  self.isa,
		                                                'organization':         self.organization,
		                                                'target.attributes':    self.compileTargetAttributes(),
		                                                'main.group.id':        self.mainGroup.id,
		                                                'configuraion.list.id': self.configurationList.id,
		                                                'products.group.id':    self.productsGroup.id,
		                                                'project.targets':      self.targets.compileList()
													} )

	# Project
	Project = """
		{id} /* Project object */ = {
			isa = {isa};
			attributes = {
				LastUpgradeCheck = 0510;
				ORGANIZATIONNAME = {organization};
				TargetAttributes = {
{target.attributes}
				};
			};
			buildConfigurationList = {configuraion.list.id} /* Build configuration list for PBXProject "{name}" */;
			compatibilityVersion = "Xcode 3.2";
			developmentRegion = English;
			hasScannedForEncodings = 0;
			knownRegions = (
				en,
			);
			mainGroup = {main.group.id};
			productRefGroup = {products.group.id} /* Products */;
			projectDirPath = "";
			projectRoot = "";
			targets = (
{project.targets}
			);
		};
"""

	# Root
	Root = """// !$*UTF8*$!
{
	archiveVersion = {archive.version};
	classes = {
	};
	objectVersion = {object.version};
	objects = {
{objects}
	};
	rootObject = {id} /* Project object */;
}
"""



#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os

from Resource       import Resource
from ObjectList     import ObjectList
from ...Template    import Template

# class Target
class Target( Resource ):
	# ctor
	def __init__( self, objects, group, type, name, path, configurations, defaultConfiguration ):
		Resource.__init__( self, 'PBXNativeTarget', name )

		self.objects            = objects
		self.group              = group
		self.generated          = None
		self.product            = self.objects.createFileReference( type, path, 'BUILT_PRODUCTS_DIR' )
		self.configurationList  = self.objects.createConfigurationList( self, configurations, defaultConfiguration )
		self.sourcePhase        = self.objects.createSourceBuildPhase( self )
		self.frameworkPhase     = self.objects.createFrameworkPhase( self )
		self.dependencies       = ObjectList( 'PBXTargetDependency' )
		self.phases             = ObjectList( 'PBXBuildPhases' )

	# compileAttributes
	def compileAttributes( self ):
		return ''

	# compile
	def compile( self ):
		self.addBuildPhases()

		return Template( Target.Root ).compile( {
													'id':                   self.id,
													'name':                 self.name,
		                                            'isa':                  self.isa,
													'configuration.list':   self.configurationList.id,
													'build.phases':         self.phases.compileList(),
		                                            'dependencies':         self.dependencies.compileList(),
		                                            'product.id':           self.product.id,
		                                            'product.fileName':     self.product.name,
		                                            'product.type':         Target.Types[self.product.type]
												} )

	# addBuildPhases
	def addBuildPhases( self ):
		self.phases.add( self.sourcePhase )
		self.phases.add( self.frameworkPhase )

	# getSourceTypeForExtension
	def getSourceTypeForExtension( self, ext ):
		types = { '.xcassets': 'folder.assetcatalog', '.h': 'sourcecode.c.h', '.c': 'sourcecode.c.c', '.cpp': 'sourcecode.cpp.cpp', '.mm': 'sourcecode.cpp.objcpp', '.plist': 'text.plist' }
		if ext in types.keys():
			return types[ext]

		return 'sourcecode'

	# resolveGroup
	def resolveGroup( self, path ):
		group = self.group

		for name in path.split( '/' ):
			if len( name ):
				group = group.resolveGroup( name )

		return group

	# addProjectLibrary
	def addProjectLibrary( self, target ):
		assert target.product != None

		# Create file reference for a product
		file = self.objects.project.addProjectLibrary( target.name )

		# Add link phase
		self.frameworkPhase.add( self.objects.createBuildFile( self, file ) )

		# Add dependency
		self.dependencies.add( self.objects.createDependency( target ) )

	# addLibrary
	def addLibrary( self, name ):
		# Create file reference for library
		file = self.objects.project.addLibrary( name )

		# Add to a link phase
		self.frameworkPhase.add( self.objects.createBuildFile( self, file ) )

		# Add library search path
		self.configurationList.forEach( lambda c: c.addSetting( 'LIBRARY_SEARCH_PATHS', os.path.dirname( name ) ) )

	# addFramework
	def addFramework( self, name ):
		# Create file reference for framework
		file = self.objects.project.addFramework( name )

		# Add link phase
		self.frameworkPhase.add( self.objects.createBuildFile( self, file ) )

		# Add framework search path for local frameworks
		if name.endswith( '.framework' ):
			self.configurationList.forEach( lambda c: c.addSetting( 'FRAMEWORK_SEARCH_PATHS', os.path.dirname( name ) ) )

	# addShellScript
	def addShellScript( self, command ):
		self.phases.add( self.objects.createShellScriptPhase( self, command ) )

	# addSourceFile
	def addSourceFile( self, fileName, parentGroup ):
		group = self.resolveGroup( parentGroup ) if len( parentGroup ) > 0 else self.group
		self.addFileToGroup( fileName, group )

	# addGeneratedFile
	def addGeneratedFile( self, fileName ):
		if not self.generated:
			self.generated = self.objects.project.addGroup( 'GeneratedFiles' )

		self.addFileToGroup( fileName, self.generated )

	# addFileToGroup
	def addFileToGroup( self, fileName, group ):
		compilable  = ['.c', '.cpp', '.m', '.mm']
		name, ext   = os.path.splitext( fileName )
		type        = self.getSourceTypeForExtension( ext )
		file        = self.objects.createFileReference( type, fileName, '"<group>"' )

		group.add( file )

		# Add a build file to source build phase
		if ext in compilable:
			self.sourcePhase.add( self.objects.createBuildFile( self, file ) )

		return file

	# Types
	Types = { 'archive.ar': 'com.apple.product-type.library.static', 'wrapper.application': 'com.apple.product-type.application', 'compiled.mach-o.executable': 'com.apple.product-type.tool' }

	# Root
	Root = """
		{id} /* {name} */ = {
			isa = {isa};
			buildConfigurationList = {configuration.list} /* Build configuration list for PBXNativeTarget "{name}" */;
			buildPhases = (
{build.phases}
			);
			buildRules = (
			);
			dependencies = (
{dependencies}
			);
			name = {name};
			productName = {name};
			productReference = {product.id} /* {product.fileName} */;
			productType = "{product.type}";
		};
"""

# class Dependency
class Dependency( Resource ):
	# ctor
	def __init__( self, objects, target ):
		Resource.__init__( self, 'PBXTargetDependency', target.name )

		self.proxy  = objects.createContainerItemProxy( target )
		self.target = target

	# compile
	def compile( self ):
		return Template( Dependency.Root ).compile( { 'id': self.id, 'isa': self.isa, 'target.id': self.target.id, 'target.name': self.target.name, 'proxy': self.proxy.id } )

	# Root
	Root = """
		{id} /* {isa} */ = {
			isa = {isa};
			target = {target.id} /* {target.name} */;
			targetProxy = {proxy} /* PBXContainerItemProxy */;
		};
"""

# class ContainerItemProxy
class ContainerItemProxy( Resource ):
	# ctor
	def __init__( self, target, project ):
		Resource.__init__( self, 'PBXContainerItemProxy', target.name )

		self.target  = target
		self.project = project

	# compile
	def compile( self ):
		return Template( ContainerItemProxy.Root ).compile( { 'id': self.id, 'isa': self.isa, 'portal': self.project.id, 'target.id': self.target.id, 'target.name': self.target.name } )

	# Root
	Root = """
		{id} /* {isa} */ = {
			isa = {isa};
			containerPortal = {portal} /* Project object */;
			proxyType = 1;
			remoteGlobalIDString = {target.id};
			remoteInfo = {target.name};
		};
"""

# class StaticLibrary
class StaticLibrary( Target ):
	# ctor
	def __init__( self, objects, group, name, path, configurations, defaultConfiguration ):
		Target.__init__( self, objects, group, 'archive.ar', name, path, configurations, defaultConfiguration )

# class Application
class Application( Target ):
	# ctor
	def __init__( self, objects, group, name, path, configurations, defaultConfiguration ):
		Target.__init__( self, objects, group, 'compiled.mach-o.executable', name, path, configurations, defaultConfiguration )

# class ApplicationBundle
class ApplicationBundle( Target ):
	# ctor
	def __init__( self, objects, group, name, path, configurations, defaultConfiguration, team ):
		Target.__init__( self, objects, group, 'wrapper.application', name, path, configurations, defaultConfiguration )
		self.team           = team
		self.resources      = None
		self.resourcePhase  = self.objects.createResourcePhase( self )

	# compileAttributes
	def compileAttributes( self ):
		return Template( ApplicationBundle.Attributes ).compile( { 'id': self.id, 'team': self.team } )

	# addBuildPhases
	def addBuildPhases( self ):
		Target.addBuildPhases( self )

		self.phases.add( self.resourcePhase )

	# addResource
	def addResource( self, file, copyToBundle ):
		# Create Resources folder
		if not self.resources:
			self.resources = self.objects.project.addGroup( 'Resources' )

		# Add file to the group
		self.resources.add( file )

		# Add file to a build phase
		if copyToBundle:
			self.resourcePhase.add( self.objects.createBuildFile( self, file ) )

	# addResourceFile
	def addResourceFile( self, name ):
		# Create file reference for icon.xcassets
		file = self.objects.createFileReference( 'file', name, '"<group>"' )

		# Add file to a group
		self.addResource( file, True )

	# addPlist
	def addPlist( self, name ):
		# Create file reference for plist
		file = self.objects.createFileReference( 'text.plist.xml', name, '"<group>"' )

		# Add file to a group
		self.addResource( file, False )

		# Set the target plist property
		self.configurationList.forEach( lambda c: c.set( 'INFOPLIST_FILE', '"{0}"'.format( name ) ) )

	# addAssetCatalog
	def addAssetCatalog( self, name ):
		# Create file reference for icon.xcassets
		file = self.objects.createFileReference( 'folder.assetcatalog', name, '"<group>"' )

		# Add file to a group
		self.addResource( file, True )

	# setIcon
	def setAppIcon( self, name ):
		self.configurationList.forEach( lambda c: c.set( 'ASSETCATALOG_COMPILER_APPICON_NAME', name ) )

	# setLaunchImage
	def setLaunchImage( self, name ):
		self.configurationList.forEach( lambda c: c.set( 'ASSETCATALOG_COMPILER_LAUNCHIMAGE_NAME', name ) )

	# Attributes
	Attributes = """
					{id} = {
						DevelopmentTeam = {team};
					};
"""


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os, uuid, codecs

from ..Generator import Generator
from ..Template  import Template

# ID
def ID():
	return '{' + str( uuid.uuid1() ).upper() + '}'

# Open
def Open( fileName ):
	return codecs.open( fileName, "w", "utf-8-sig" )

# class Windows
class Windows( Generator ):
	# constructor
	def __init__( self ):
		Generator.__init__( self )

		self.id 		= ID()
		self.projects 	= {}

	# generate
	def generate( self ):
		print( 'Generating Win32 project...' )
		Generator.generate( self )

		self.generateSolutionProjects()
		self.generateSolution()
		
	# generateSolution
	def generateSolution( self ):
		file = Open( os.path.join( self.binaryDir, self.sourceProject.name + '.sln' ) )
		file.write( Template( Windows.Solution ).compile( 	{
																	'projects':			self.compileSolutionProjects(),
																	'configurations':	self.compileSolutionConfigurations(),
																}, True ) )
		file.close()

	# generateSolutionProjects
	def generateSolutionProjects( self ):
		# Project
		class Project:
			def __init__( self, name, path ):
				self.id   		 = ID()
				self.name 		 = name
				self.files		 = {}
				self.fileName	 = os.path.join( path, name + '.vcxproj' )
				self.filtersFile = self.fileName + '.filters'

		# addTargetProject
		def addTargetProject( name, target ):
			path = self.getPathForTarget( target )
			self.projects[name] = Project( name, os.path.relpath( path, self.binaryDir ) )

		# Generate project list
		self.forEachTarget( addTargetProject )

		# Generate project file
		for name in self.projects:
			project = self.projects[name]
			
			# Add proxy files for commands
			self.generateProxyFiles( project )

			# Generate
			file = Open( os.path.join( self.binaryDir, project.fileName ) )
			file.write( Template( Windows.Project ).compile( 	{
																		'configuration.list': 	Windows.ProjectConfigurationList,
																		'id':					project.id,
																		'name':					project.name,
																		'configurations':		self.generateConfigurations( project ),
																		'settings':				self.generateSettings( project ),
																		'commands':				self.generateCommands( project ),
																		'items':				self.generateItems( project ),
																	}, True ) )
			file.close()

			file = Open( os.path.join( self.binaryDir, project.filtersFile ) )
			file.write( Template( Windows.Filters ).compile( 	{
																		'groups': 	self.generateFilterGroups( project ),
																		'compile':	self.generateFilterFiles( project, Windows.FilterCompileItem, ['.c', '.cpp'] ),
																		'headers':	self.generateFilterFiles( project, Windows.FilterHeaderItem, ['.h'] ),
																		'commands':	self.generateFilterFiles( project, Windows.FilterRuleItem, ['.rule'] ),
																	}, True ) )
			file.close()

	# generateFilterGroups
	def generateFilterGroups( self, project ):
		target = self.findTargetByName( project.name )

		# generateFilterGroup
		def generateFilterGroup( target, path, files ):
			return Template( Windows.FilterGroup ).compile( { 'path': path, 'id': ID() }, True )

		return self.processEachGroup( target, generateFilterGroup )

	# generateFilterFiles
	def generateFilterFiles( self, project, tpl, filter ):
		target = self.findTargetByName( project.name )

		# generateFilterItems
		def generateFilterGroup( target, path, files ):
			result = ''

			for fileName in files:
				name, ext = os.path.splitext( fileName )
				if ext in filter:
					result += Template( tpl ).compile( { 'fileName': fileName, 'path': path }, True )

			return result

		return self.processEachGroup( target, generateFilterGroup )

	# def generateProjectLinker
	def generateProjectLinker( self, project, debug ):
		target  = self.findTargetByName( project.name )

		if target.type != 'executable' and target.type != 'shared':
			return ''

		# generateLibs
		def generateLibs( target, name, lib ):
			return name + '.lib;'

		options  	= ' /machine:X86 ' + ('/debug ' if debug else '')
		output 	 	= os.path.join( self.binaryDir, '$(Configuration)' )
		libs     	= self.processEachTargetLib( target, generateLibs )
		library  	= os.path.join( output, project.name + '.lib' )
		pdb 	 	= os.path.join( output, project.name + '.pdb' )
		subsystem 	= 'Windows'
		
		if 'subsystem' in target.params.keys():
			subsystem = 'Console' if target.params['subsystem'] == 'console' else 'Windows'
		
		# Add Qt to library paths
		if debug:
			libs = libs.replace( 'QtCore', 'QtCored4' ).replace( 'QtGui', 'QtGuid4' ).replace( 'QtOpenGL', 'QtOpenGLd4' )
		else:
			libs = libs.replace( 'QtCore', 'QtCore4' ).replace( 'QtGui', 'QtGui4' ).replace( 'QtOpenGL', 'QtOpenGL4' )

		if libs.find( 'Qt' ) != -1:
			Qt = self.makefile.get( 'Qt' )

			if Qt == None or not os.path.exists( Qt ):
				print 'ERROR: no Qt path set'
			else:
				output += ';' + os.path.join( Qt, 'lib' )

		return Template( Windows.Link ).compile( 	{
															'libs': 			libs,
															'lib.path':			output,
															'options': 			options,
															'is.debug':			'true' if debug else 'false',
															'subsystem':		subsystem,
															'import.library': 	library,
															'pdb':				pdb,
														}, True )

	# generateConfigurations
	def generateConfigurations( self, project ):
		result = ''
		type   = { 'shared': 'DynamicLibrary', 'static': 'StaticLibrary', 'executable': 'Application' }
		target = self.findTargetByName( project.name )

		result += Template( Windows.ConfigurationPropertyGroup ).compile( { 'configuration': 'Debug|Win32', 'type': type[target.type] }, True )
		result += Template( Windows.ConfigurationPropertyGroup ).compile( { 'configuration': 'Release|Win32', 'type': type[target.type] }, True )

		return result

	# generateProxyFiles
	def generateProxyFiles( self, project ):
		target = self.findTargetByName( project.name )

		# generateCommand
		def generateCommand( target, cmd ):
			fileName = os.path.join( self.getPathForTarget( target ), os.path.basename( cmd.output ) + '.rule' )
			
			fp = open( fileName, 'wt' )
			fp.write( '# Generated by YAP' )
			fp.close()
			
			target.files( fileName )
			target.group( 'GeneratedFiles', [fileName] )

		self.processEachTargetCommand( target, generateCommand )		

	# generateCommands
	def generateCommands( self, project ):
		result = ''
		target = self.findTargetByName( project.name )

		# generateCommand
		def generateCommand( target, cmd ):
			input   = ''
			output  = ''
			message = 'Generating'

			for ext in cmd.generatedExtensions:
				output  += cmd.output + ext + ';'
				message += ' ' + os.path.basename( cmd.output + ext )
				
			message += '...'

			path 	 = self.getPathForTarget( target )
			fileName = os.path.join( path, os.path.basename( cmd.output ) )
			input 	 = os.path.relpath( fileName + '.rule', path )
			
			header  = Template( Windows.CommandHeader ).compile( { 'input': input } )
			command = Template( Windows.Command ).compile( { 'command': cmd.command, 'message': message, 'input': input, 'output': output, 'additional.input': cmd.input[0] }, True )
			footer  = Template( Windows.CommandFooter ).compile( {} )

			return header + command + footer

		return self.processEachTargetCommand( target, generateCommand )

	# generateSettings
	def generateSettings( self, project ):
		result = ''

		includes 	 = ''
		preprocessor = ''
		target   	 = self.findTargetByName( project.name )
		path 	 	 = self.getPathForTarget( target )

		def generateInclude( target, path ):
			return path + ';'

		def generateDefine( target, define ):
			return define + ';'

		includes 	 = self.processEachTargetInclude( target, generateInclude )
		preprocessor = self.processEachTargetDefine( target, generateDefine )

		# Debug
		result += Template( Windows.ProjectConfiguration ).compile( {
																			'configuration': 	'Debug|Win32',
																			'inline.function':	'Disabled',
																			'debug.format':		'ProgramDatabase',
																			'optimization': 	'Disabled',
																			'includes':			includes,
																			'preprocessor':		'_DEBUG;' + preprocessor,
																			'runtime.lib':		'MultiThreadedDebugDLL',
																			'link':				self.generateProjectLinker( project, True ),
																		 }, True )

		# Release
		result += Template( Windows.ProjectConfiguration ).compile( {
																			'configuration': 	'Release|Win32',
																			'inline.function':	'AnySuitable',
																			'debug.format':		'',
																			'optimization': 	'MaxSpeed',
																			'includes':			includes,
																			'preprocessor':		preprocessor,
																			'runtime.lib':		'MultiThreadedDLL',
																			'link':				self.generateProjectLinker( project, False ),
																		 }, True )

		return result

	# generateItems
	def generateItems( self, project ):
		result 	= ''
		target 	= self.findTargetByName( project.name )
		path 	= self.getPathForTarget( target )
		files	= project.files

		# Files
		for fileName in target.sources:
			name, ext 	= os.path.splitext( os.path.basename( fileName ) )
			filePath  	= os.path.relpath( fileName, path )
			objFileName = name.upper() + ext + '.obj'
			
			# Generate valid obj file name
			if objFileName in files.keys():
				files[objFileName] = files[objFileName] + 1
				objFileName 	   += str( files[objFileName] )
				objFilePath		   = '<ObjectFileName>$(Configuration)\\{0}</ObjectFileName>'.format( objFileName )
			else:
				objFilePath		   = ''
				files[objFileName] = 1

			if ext == '.h':
				result += Template( Windows.ItemHeader ).compile( { 'fileName': filePath }, True )
			elif ext == '.cpp':
				result += Template( Windows.ItemCpp ).compile( { 'fileName': filePath, 'obj.fileName': objFilePath }, True )
			elif ext == '.c':
				result += Template( Windows.ItemC ).compile( { 'fileName': filePath, 'obj.fileName': objFilePath }, True )

		return result
		
	################################### TEMPLATE COMPILERS
		
	# compileSolutionProjects
	def compileSolutionProjects( self ):
		result 		 = ''

		for name in self.projects:
			project 	 = self.projects[name]
			target 	 	 = self.findTargetByName( name )
			dependencies = ''
			
			for lib in target.libs:
				if lib in self.sourceProject.targets.keys():
					dependencies += Template( Windows.SolutionDependency ).compile( { 'uuid': self.projects[lib].id }, True )
			
			result += Template( Windows.SolutionProject ).compile( {
																 			'solution.id': 	self.id,
																 			'name': 		project.name,
																 			'id':			project.id,
																 			'fileName':		project.fileName,
																			'dependencies':	dependencies,
																		}, True )

		return result
			
	# compileSolutionConfigurations
	def compileSolutionConfigurations( self ):
		result = ''

		for name in self.projects:
			result += Template( Windows.SolutionProjectConfigurations ).compile( { 'id': self.projects[name].id }, True )

		return result
		
	################################### TEMPLATES
	
	Solution = """
Microsoft Visual Studio Solution File, Format Version 11.00
# Visual Studio 2010
{projects}
Global
	GlobalSection(SolutionConfigurationPlatforms) = preSolution
		Debug|Win32 = Debug|Win32
		Release|Win32 = Release|Win32
	EndGlobalSection
	GlobalSection(ProjectConfigurationPlatforms) = postSolution{configurations}
	EndGlobalSection
	GlobalSection(SolutionProperties) = preSolution
		HideSolutionNode = FALSE
	EndGlobalSection
EndGlobal
"""

	SolutionProject = """
Project("{solution.id}") = "{name}", "{fileName}", "{id}"
{dependencies}
EndProject"""

	SolutionDependency = """
	ProjectSection(ProjectDependencies) = postProject
		{uuid} = {uuid}
	EndProjectSection
"""

	SolutionProjectConfigurations = """
		{id}.Debug|Win32.ActiveCfg = Debug|Win32
		{id}.Debug|Win32.Build.0 = Debug|Win32
		{id}.Release|Win32.ActiveCfg = Release|Win32
		{id}.Release|Win32.Build.0 = Release|Win32"""

	ProjectProperty = """
<PropertyGroup Condition="'$(Configuration)|$(Platform)'=='{configuration}'" Label="Configuration">
    <ConfigurationType>{type}</ConfigurationType>
    <UseOfMfc>false</UseOfMfc>
    <CharacterSet>MultiByte</CharacterSet>
</PropertyGroup>
"""

	ProjectConfiguration = """
<ItemDefinitionGroup Condition="'$(Configuration)|$(Platform)'=='{configuration}'">
  <ClCompile>
    <AdditionalOptions>/MP8 /Zm1000 %(AdditionalOptions)</AdditionalOptions>
    <AdditionalIncludeDirectories>{includes}%(AdditionalIncludeDirectories)</AdditionalIncludeDirectories>
    <CompileAs>CompileAsCpp</CompileAs>
    <ExceptionHandling>Sync</ExceptionHandling>
    <InlineFunctionExpansion>{inline.function}</InlineFunctionExpansion>
    <Optimization>{optimization}</Optimization>
    <RuntimeLibrary>{runtime.lib}</RuntimeLibrary>
    <RuntimeTypeInfo>true</RuntimeTypeInfo>
    <WarningLevel>Level3</WarningLevel>
    <DebugInformationFormat>{debug.format}</DebugInformationFormat>
    <PreprocessorDefinitions>WIN32;_WINDOWS;_CRT_SECURE_NO_WARNINGS;{preprocessor}%(PreprocessorDefinitions)</PreprocessorDefinitions>
    <AssemblerListingLocation>Release</AssemblerListingLocation>
    <ObjectFileName>$(IntDir)</ObjectFileName>
  </ClCompile>
  <ResourceCompile>
    <PreprocessorDefinitions>WIN32;_WINDOWS;_CRT_SECURE_NO_WARNINGS;{preprocessor}%(PreprocessorDefinitions)</PreprocessorDefinitions>
    <AdditionalIncludeDirectories>{includes}%(AdditionalIncludeDirectories)</AdditionalIncludeDirectories>
  </ResourceCompile>
  <Midl>
    <AdditionalIncludeDirectories>{includes}%(AdditionalIncludeDirectories)</AdditionalIncludeDirectories>
    <OutputDirectory>$(IntDir)</OutputDirectory>
    <HeaderFileName>%(Filename).h</HeaderFileName>
    <TypeLibraryName>%(Filename).tlb</TypeLibraryName>
    <InterfaceIdentifierFileName>%(Filename)_i.c</InterfaceIdentifierFileName>
    <ProxyFileName>%(Filename)_p.c</ProxyFileName>
  </Midl>
  {link}
</ItemDefinitionGroup>
"""

	Link = """
  <Link>
    <AdditionalLibraryDirectories>{lib.path};%(AdditionalLibraryDirectories)</AdditionalLibraryDirectories>
    <AdditionalOptions>{options} %(AdditionalOptions)</AdditionalOptions>
    <AdditionalDependencies>kernel32.lib;user32.lib;gdi32.lib;winspool.lib;shell32.lib;ole32.lib;oleaut32.lib;uuid.lib;comdlg32.lib;advapi32.lib;{libs}</AdditionalDependencies>
    <GenerateDebugInformation>{is.debug}</GenerateDebugInformation>
    <ImportLibrary>{import.library}</ImportLibrary>
    <LinkIncremental>{is.debug}</LinkIncremental>
    <ProgramDataBaseFileName>{pdb}</ProgramDataBaseFileName>
    <StackReserveSize>10000000</StackReserveSize>
    <SubSystem>{subsystem}</SubSystem>
    <Version></Version>
  </Link>
	"""

	Project = """<?xml version="1.0" encoding="utf-8"?>
<Project DefaultTargets="Build" ToolsVersion="4.0" xmlns="http://schemas.microsoft.com/developer/msbuild/2003">
  <ItemGroup Label="ProjectConfigurations">
    {configuration.list}
  </ItemGroup>
  <PropertyGroup Label="Globals">
    <ProjectGuid>{id}</ProjectGuid>
    <Platform>Win32</Platform>
    <ProjectName>{name}</ProjectName>
  </PropertyGroup>
  <Import Project="$(VCTargetsPath)\Microsoft.Cpp.Default.props" />
  {configurations}
  <Import Project="$(VCTargetsPath)\Microsoft.Cpp.props" />
  <ImportGroup Label="ExtensionSettings">
  </ImportGroup>
  <ImportGroup Label="PropertySheets" Condition="'$(Configuration)|$(Platform)'=='Debug|Win32'">
    <Import Project="$(UserRootDir)\Microsoft.Cpp.$(Platform).user.props" Condition="exists('$(UserRootDir)\Microsoft.Cpp.$(Platform).user.props')" Label="LocalAppDataPlatform" />
  </ImportGroup>
  <ImportGroup Label="PropertySheets" Condition="'$(Configuration)|$(Platform)'=='Release|Win32'">
    <Import Project="$(UserRootDir)\Microsoft.Cpp.$(Platform).user.props" Condition="exists('$(UserRootDir)\Microsoft.Cpp.$(Platform).user.props')" Label="LocalAppDataPlatform" />
  </ImportGroup>

  {settings}

  {commands}

  <ItemGroup>
  	{items}
  </ItemGroup>

  <Import Project="$(VCTargetsPath)\Microsoft.Cpp.targets" />
  <ImportGroup Label="ExtensionTargets">
  </ImportGroup>
</Project>
"""

	Filters = """<?xml version="1.0" encoding="utf-8"?>
<Project ToolsVersion="4.0" xmlns="http://schemas.microsoft.com/developer/msbuild/2003">
  <ItemGroup>
    {compile}
  </ItemGroup>
  <ItemGroup>
    {headers}
  </ItemGroup>
  <ItemGroup>
	{commands}
  </ItemGroup>
  <ItemGroup>
    {groups}
  </ItemGroup>
</Project>
"""

	ProjectConfigurationList = """
<ProjectConfiguration Include="Debug|Win32">
  <Configuration>Debug</Configuration>
  <Platform>Win32</Platform>
</ProjectConfiguration>
<ProjectConfiguration Include="Release|Win32">
  <Configuration>Release</Configuration>
  <Platform>Win32</Platform>
</ProjectConfiguration>
"""

	ConfigurationPropertyGroup = """
<PropertyGroup Condition="'$(Configuration)|$(Platform)'=='{configuration}'" Label="Configuration">
  <ConfigurationType>{type}</ConfigurationType>
  <UseOfMfc>false</UseOfMfc>
  <CharacterSet>MultiByte</CharacterSet>
</PropertyGroup>
"""

	ItemHeader = '<ClInclude Include="{fileName}" />\n'
	ItemCpp    = """<ClCompile Include="{fileName}" >
		{obj.fileName}
	</ClCompile>"""
	ItemC 	   = """<ClCompile Include="{fileName}">
	  {obj.fileName}
      <CompileAs Condition="'$(Configuration)|$(Platform)'=='Debug|Win32'">CompileAsC</CompileAs>
      <CompileAs Condition="'$(Configuration)|$(Platform)'=='Release|Win32'">CompileAsC</CompileAs>
    </ClCompile>"""

	FilterGroup = """<Filter Include="{path}">
  <UniqueIdentifier>{id}</UniqueIdentifier>
</Filter>
    """

	FilterCompileItem = """<ClCompile Include="{fileName}">
  <Filter>{path}</Filter>
</ClCompile>
    """

	FilterHeaderItem = """<ClInclude Include="{fileName}">
  <Filter>{path}</Filter>
</ClInclude>
    """
	
	FilterRuleItem = """<CustomBuild Include="{fileName}">
  <Filter>{path}</Filter>
</CustomBuild>
    """
	
	CommandHeader = """
	<ItemGroup>
	<CustomBuild Include="{input}">
"""

	CommandFooter = """
	</CustomBuild>
	</ItemGroup>
"""

	Command = """
    <Message>{message}</Message>
    <Command>{command}</Command>
    <AdditionalInputs>{additional.input};%(AdditionalInputs)</AdditionalInputs>
    <Outputs>{output}</Outputs>
"""


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os

from Resource    import Resource
from ...Template import Template

# class FileReference
class FileReference( Resource ):
	# ctor
	def __init__( self, type, path, dir ):
		Resource.__init__( self, 'PBXFileReference', os.path.basename( path ) )
		self.type = type
		self.path = path
		self.dir  = dir

		# static library - archive.ar

	# compile
	def compile( self ):
		return Template( FileReference.Root ).compile( { 'id': self.id, 'isa': self.isa, 'name': self.name, 'type': self.type, 'path': self.path, 'dir': self.dir } )

	# Root
	Root = "\t\t{id} /* {name} */ = {isa = {isa}; explicitFileType = {type}; includeInIndex = 0; name = {name}; path = {path}; sourceTree = {dir}; };\n"

# class BuildFile
class BuildFile( Resource ):
	# ctor
	def __init__( self, owner, file ):
		Resource.__init__( self, 'PBXBuildFile', file.name )
		self.file  = file
		self.owner = owner

	# compile
	def compile( self ):
		return Template( BuildFile.Root ).compile( { 'id': self.id, 'isa': self.isa, 'owner': self.owner.name, 'file.name': self.file.name, 'file.reference': self.file.id } )

	# Root
	Root = "\t\t{id} /* {file.name} in {owner} */ = {isa = {isa}; fileRef = {file.reference} /* {file.name} */; };\n"


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

from ..Generator import Generator
from Make        import Make
from ..Template  import Template

# class Flash
class Flash( Make ):
	# constructor
	def __init__( self ):
		Make.__init__( self )
		self.extensions = { 'shared': 'swc', 'static': 'a', 'executable': 'swf' }

	# generate
	def generate( self ):
		self.toolchain  = { 'cc': Generator.FLACC + '/sdk/usr/bin/gcc', 'cxx': Generator.FLACC + '/sdk/usr/bin/g++', 'ar': Generator.FLACC + '/sdk/usr/bin/ar' }
		
		print( 'Generating Flash project...' )
		MakefileProject.generate( self )

		Template( Flash.Exports ).compileToFile( Generator.BinaryDir + '/exports.txt' )

	# getFlagsForTarget
	def getFlagsForTarget( self, target ):
		flags = MakefileProject.getFlagsForTarget( self, target )
		return '-O4 -flto-api=../exports.txt -w -Wno-write-strings -Wno-trigraphs -fno-rtti -jvmopt=-Xmx900M ' + flags

	# getProductForTarget
	def getProductForTarget( self, target, forLinker = False ):
		ext  = self.getExtensionForTarget( target )
		name = target.name

		if len( ext ) == 0:
			return name

		if not forLinker:
			name = 'lib' + name

		return name + '.' + ext

	# getLinkerFlagsForTarget
	def getLinkerFlagsForTarget( self, target ):
		libs = self.getLibrariesForTarget( target )

		if target.type == 'shared':
			return '{0} -O4 -jvmopt=-Xmx1G -flto-api=../exports.txt -lAS3++ -lFlash++ -emit-swc=native.{1}'.format( libs, target.name )
		elif target.type == 'executable':
			return '{0} -O4 -jvmopt=-Xmx1G -flto-api=../exports.txt -lAS3++ -lFlash++ -symbol-abc=Console.abc -emit-swf -swf-size=640x480'.format( libs )

		return ''

	Exports = """
# built in symbols that must always be preserved
_start1
__muldi3
__divdi3
malloc
free
memcpy
memmove
flascc_uiTickProc
_sync_synchronize

# symbols for C++ exception handling
_Unwind_SjLj_Register
_Unwind_SjLj_Resume
_Unwind_SjLj_Unregister
_Unwind_SjLj_RaiseException
"""


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

from Resource       import Resource
from ObjectList     import ObjectList
from ...Template    import Template

# class ConfigurationList
class ConfigurationList( Resource ):
	# ctor
	def __init__( self, default, owner ):
		Resource.__init__( self, 'XCConfigurationList', None )

		self.default = default
		self.owner   = owner
		self.items   = ObjectList()

	# add
	def add( self, item ):
		self.items.add( item )

	# forEach
	def forEach( self, callback ):
		for item in self.items.items:
			callback( item )

	# compile
	def compile( self ):
		return Template( ConfigurationList.Root ).compile( { 'id': self.id, 'isa': self.isa, 'items': self.items.compileList(), 'default': self.default, 'owner': self.owner.name, 'owner.isa': self.owner.isa } )

	# Root
	Root = """
		{id} /* Build configuration list for {owner.isa} {owner} */ = {
			isa = {isa};
			buildConfigurations = (
{items}
			);
			defaultConfigurationIsVisible = 0;
			defaultConfigurationName = {default};
		};
"""


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os

# class RTTI
class RTTI:
	# ctor
	def __init__( self, target, makefile ):
		self.makefile = makefile

	# wrap
	def wrap( self, target, generator, fileName, extension ):
		if not os.path.exists( fileName ) or extension != '.h':
			return False

		# Check if we have RTTI macroses
		fh   = open( fileName )
		data = fh.read()
		fh.close()

		# Search for RTTI macroses
		macroses 	= [ 'dcBeginClass', 'dcBeginNetworkPacket', 'dcBeginSerializable', 'dcBeginSerializableSuper', 'dcBeginEnum' ]
		hasMacroses = False

		for macro in macroses:
			if data.find( macro ) != -1:
				hasMacroses = True

		if not hasMacroses or data.find( '#skip.rtti' ) != -1:
			return

		# Add command
		name, ext 	 	= os.path.splitext( os.path.basename( fileName ) )
		outputFolder	= generator.getPathForTarget( target )
		outputFileName	= os.path.join( outputFolder, 'rtti' + name )
		cmd 			= 'python ' + os.path.join( generator.sourceDir, '../tools/scripts/rtti.py' ) + ' ' + fileName + ' ' + outputFolder + '/rtti' + name + '.cpp'

		target.command( [ fileName ], outputFileName, cmd, 'Generating rtti{0}.cpp...'.format( name ), ['.cpp'] )

		return True

	#	target.message( 'Has RTTI for ' + name )



#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

from Resource    import Resource
from ...Template import Template

# class BuildConfiguration
class BuildConfiguration( Resource ):
	# ctor
	def __init__( self, name, settings ):
		Resource.__init__( self, 'XCBuildConfiguration', name )
		self.settings = settings

	# addSetting
	def addSetting( self, key, value ):
		# Add the key if doesn't exist
		if not key in self.settings.keys():
			self.settings[key] = []

		# Check for duplicates and add
		if not value in self.settings[key]:
			self.settings[key].append( value )

	# set
	def set( self, key, value ):
		self.settings[key] = value

	# compileSettings
	def compileSettings( self ):
		result = ''

		for k, v in self.settings.items():
			if type( v ) == list:
				value = '(\n'
				for i in v:
					value += '\t\t\t\t\t"' + i + '",\n'
				value += '\t\t\t\t)'
			else:
				value = v

			result += Template( BuildConfiguration.Setting ).compile( { 'key': k, 'value': value } )

		return result

	# compile
	def compile( self ):
		return Template( BuildConfiguration.Root ).compile( { 'id': self.id, 'isa': self.isa, 'name': self.name, 'settings': self.compileSettings() } )

	# Setting
	Setting = "\t\t\t\t{key} = {value};\n"

	# Root
	Root = """
		{id} /* {name} */ = {
			isa = {isa};
			buildSettings = {
{settings}
			};
			name = {name};
		};
"""



#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import glob, os

from File import File

# class Folder
class Folder:
	# ctor
	def __init__( self, target, name = '', parent = None ):
		self._name      = name
		self._parent    = parent
		self._target    = target
		self._files     = []
		self._folders   = {}

	# sourcePath
	@property
	def sourcePath( self ):
		return os.path.join( self._parent.sourcePath, self._name ) if self._parent else self._name

	# addFile
	def addFile( self, fileName ):
		self._files.append( File( self._target, self, fileName ) )

	# addFileAtPath
	def addFileAtPath( self, path ):
		self.resolve( path ).addFile( os.path.basename( path ) )

	# addFilesFromDirectory
	def addFilesFromDirectory( self, path ):
		if path.endswith( '/*' ):
			return self.addFilesFromDirectoryRecursive( path )

		for fileName in glob.glob( os.path.join( path, '*.*' ) ):
			self.addFileAtPath( os.path.relpath( fileName, self._target.sourcePath ) )

	# addFilesFromDirectoryRecursive
	def addFilesFromDirectoryRecursive( self, path ):
		self.addFilesFromDirectory( path.replace( '/*', '/' ) )

		# Recursively add all nested folders
		for folder in glob.glob( path ):
			if os.path.isdir( folder ):
				# Continue recursive add
				self.addFilesFromDirectoryRecursive( os.path.join( folder, '*' ) )

	# resolve
	def resolve( self, path ):
		items = path.split( '/' )
		name  = items[0]

		if len( items ) == 1:
			return self

		if not name in self._folders.keys():
			self._folders[name] = Folder( self._target, name, self )

		return self._folders[name].resolve( '/'.join( items[1:] ) )

	# filterFiles
	def filterFiles( self, filter ):
		result = [file for file in self._files if filter == None or filter( file )]

		for name, folder in self._folders.items():
			result = result + folder.filterFiles( filter )

		return result


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os
import sys

from ..Makefile  import Makefile
from Target      import Target

sys.path.insert( 1, os.path.join( sys.path[0], '..' ) )

# class Project
class Project( Target ):
	# ctor
	def __init__( self, name, platform, importer, generator ):
		Target.__init__( self, name )

		self.message( 'Configuring build environment for ' + name + '...' )

		self._targets 		= []
		self.importer		= importer
		self.platform       = platform
		self.generator      = generator
		self.toolsPath 		= None
		self.outputLibPath	= None
		self.outputExePath	= None

	# target
	def target( self, *list ):
		for item in list:
			current = Makefile.getCurrentSourceDir()
			binary  = Makefile.getCurrentBinaryDir()

			Makefile.CurrentSourceDir.append( os.path.join( current, item ) )
			Makefile.CurrentBinaryDir.append( os.path.join( binary, item + '.dir' ) )

			self.importer( os.path.join( current, item, 'Makefile.py' ) )

			Makefile.CurrentBinaryDir.pop()
			Makefile.CurrentSourceDir.pop()

	# registerTarget
	def registerTarget( self, target ):
		self._targets.append( target )

	# setToolsPath
	def setToolsPath( self, path ):
		self.toolsPath = path
		
	# setLibraryOutput
	def setLibraryOutput( self, path ):
		self.outputLibPath = path

	# setExecutableOutput
	def setExecutableOutput( self, path ):
		self.outputExePath = path

	# filterTargets
	def filterTargets( self, filter = None ):
		return [target for target in self._targets if filter == None or filter( target )]

	# findTarget
	def findTarget( self, name, types ):
		for target in self._targets:
			if target.name == name and target.type in types:
				return target

		return None


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os, glob, shutil

from ..Generator import Generator
from ..Template  import Template
from ..          import Make

# class Android
class Android( Generator ):
	# ctor
	def __init__( self ):
		Generator.__init__( self )

	# getPathForTarget
	def getPathForTarget( self, target ):
		return self.binaryDir + '/jni/' + target.name + '.dir'

	# generate
	def generate( self ):
		ndk = Make.Android()
		ndk.initialize( self.sourceDir, self.binaryDir, self.sourceProject )
		ndk.generate()

		print( 'Generating Android Eclipse project...' )

		# generateJavaSourceLinks
		def generateJavaSourceLinks( name, target ):
			result = ''

			for package in target.javaPackages:
				name    = os.path.basename( package )
				result += Template( Android.JavaSource ).compile( { 'name': name, 'path': package } )

			return result

		# generateJavaSourcesClassPaths
		def generateJavaSourcesClassPaths( name, target ):
			result = ''

			for package in target.javaPackages:
				result += '<classpathentry kind="src" path="{0}"/>'.format( os.path.basename( package ) )

			return result

		minsdk    = 9
		targetsdk = 17
		appname   = 'Player'
		package   = 'com.yam'
		activity  = '.StartupActivity'
		sources   = self.processEachTarget( generateJavaSourceLinks )
		classes   = self.processEachTarget( generateJavaSourcesClassPaths )
		src       = os.path.join( self.binaryDir, 'src' )
		gen       = os.path.join( self.binaryDir, 'gen' )
		resvalues = os.path.join( self.binaryDir, 'res/values' )

		# Create folders
		if not os.path.exists( src ):       os.makedirs( src )
		if not os.path.exists( gen ):       os.makedirs( gen )
		if not os.path.exists( resvalues ): os.makedirs( resvalues )

		# Icons
		self.addIcons( self.sourceProject.targets['Player'] )

		# Assets
		self.addAssets( self.sourceProject.targets['Player'] )

		# Strings
		Template( Android.Strings ).compileToFile( os.path.join( resvalues, 'strings.xml' ), { 'name': appname } )

		# Styles
		Template( Android.Styles ).compileToFile( os.path.join( resvalues, 'styles.xml' ) )

		# Manifest
		Template( Android.Manifest ).compileToFile( os.path.join( self.binaryDir, 'AndroidManifest.xml' ),
		                                            { 'package': package, 'min.sdk': minsdk, 'target.sdk': targetsdk, 'activity': activity, 'name': appname } )

		# Classpath
		Template( Android.Classpath ).compileToFile( os.path.join( self.binaryDir, '.classpath' ), { 'java.sources': classes } )

		# Properties
		Template( Android.Properties ).compileToFile( os.path.join( self.binaryDir, 'project.properties' ), { 'target.sdk': targetsdk } )

		# Project
		Template( Android.Project ).compileToFile( os.path.join( self.binaryDir, '.project' ), { 'name': appname, 'java.sources': sources } )

		# C Project
		Template( Android.CProject ).compileToFile( os.path.join( self.binaryDir, '.cproject' ), { 'name': appname } )

	# addIcons
	def addIcons( self, target ):
		if not target.resources:
			return

		icons = os.path.join( target.sourcePath, target.resources, 'images', 'icons.android', '*.png' )

		for icon in glob.glob( icons ):
			name, ext = os.path.splitext( os.path.basename( icon ) )
			dst       = os.path.join( self.binaryDir, 'res/drawable-' + name )

			# Remove previous folder
			if os.path.exists( dst ):
				shutil.rmtree( dst )

			# Create folder
			os.makedirs( dst )

			shutil.copyfile( icon, os.path.join( dst, 'icon.png' ) )

	# addAssets
	def addAssets( self, target ):
		if not target.resources:
			return

		assets = os.path.join( target.sourcePath, target.resources, 'assets' )

		if os.path.exists( assets ):
			dst       = os.path.join( self.binaryDir, 'assets' )

			# Remove previous folder
			if os.path.exists( dst ):
				shutil.rmtree( dst )

			shutil.copytree( assets, dst )

	# Manifest
	Manifest = """<manifest xmlns:android="http://schemas.android.com/apk/res/android"
    package="{package}"
    android:versionCode="1"
    android:versionName="1.0" >

    <uses-sdk android:minSdkVersion="{min.sdk}" android:targetSdkVersion="{target.sdk}" />
    <uses-permission android:name="android.permission.INTERNET" />
    <uses-permission android:name="android.permission.ACCESS_NETWORK_STATE"/>
    <uses-permission android:name="android.permission.ACCESS_WIFI_STATE"/>
    <uses-permission android:name="android.permission.WRITE_EXTERNAL_STORAGE"/>
    <uses-permission android:name="android.permission.READ_EXTERNAL_STORAGE"/>
    <uses-permission android:name="android.permission.READ_PHONE_STATE"/>
    <uses-permission android:name="com.android.vending.BILLING" />

    <application android:allowBackup="true" android:icon="@drawable/icon" android:label="@string/AppName" android:theme="@style/AppDefaultTheme" >
		<activity android:name="{activity}" android:label="@string/AppName" android:configChanges="orientation|keyboardHidden" android:screenOrientation="landscape">
            <meta-data android:name="android.app.lib_name" android:value="{name}" />
			<intent-filter>
				<action android:name="android.intent.action.MAIN" />
				<category android:name="android.intent.category.LAUNCHER" />
			</intent-filter>
        </activity>
    </application>
</manifest>"""

	Classpath = """<?xml version="1.0" encoding="UTF-8"?>
<classpath>
{java.sources}
	<classpathentry kind="src" path="gen"/>
	<classpathentry kind="con" path="com.android.ide.eclipse.adt.ANDROID_FRAMEWORK"/>
	<classpathentry exported="true" kind="con" path="com.android.ide.eclipse.adt.LIBRARIES"/>
	<classpathentry kind="con" path="com.android.ide.eclipse.adt.DEPENDENCIES"/>
	<classpathentry kind="output" path="bin/classes"/>
</classpath>"""

	Properties = """# Autogenerated by YAP build tool
target=android-{target.sdk}"""

	Strings = """<resources>
    <string name="AppName">{bundle.name}</string>
</resources>"""

	Styles = """<resources>
    <!--
        Base application theme, dependent on API level. This theme is replaced
        by AppBaseTheme from res/values-vXX/styles.xml on newer devices.
    -->
    <style name="AppBaseTheme" parent="android:Theme.Light">
        <!--
            Theme customizations available in newer API levels can go in
            res/values-vXX/styles.xml, while customizations related to
            backward-compatibility can go here.
        -->
    </style>

    <!-- Application theme. -->
    <style name="AppDefaultTheme" parent="AppBaseTheme">
        <!-- All customizations that are NOT specific to a particular API-level can go here. -->
    </style>
</resources>"""

	Project = """<?xml version="1.0" encoding="UTF-8"?>
<projectDescription>
	<name>{name}</name>
	<comment></comment>
	<projects>
	</projects>
	<buildSpec>
		<buildCommand>
			<name>org.eclipse.cdt.managedbuilder.core.genmakebuilder</name>
			<triggers>clean,full,incremental,</triggers>
			<arguments>
				<dictionary>
					<key>?children?</key>
					<value>?name?=outputEntries\|?children?=?name?=entry\\\\\\\|\\\|?name?=entry\\\\\\\|\\\|\||</value>
				</dictionary>
				<dictionary>
					<key>?name?</key>
					<value></value>
				</dictionary>
				<dictionary>
					<key>org.eclipse.cdt.make.core.append_environment</key>
					<value>true</value>
				</dictionary>
				<dictionary>
					<key>org.eclipse.cdt.make.core.buildArguments</key>
					<value></value>
				</dictionary>
				<dictionary>
					<key>org.eclipse.cdt.make.core.buildCommand</key>
					<value>ndk-build</value>
				</dictionary>
				<dictionary>
					<key>org.eclipse.cdt.make.core.cleanBuildTarget</key>
					<value>clean</value>
				</dictionary>
				<dictionary>
					<key>org.eclipse.cdt.make.core.contents</key>
					<value>org.eclipse.cdt.make.core.activeConfigSettings</value>
				</dictionary>
				<dictionary>
					<key>org.eclipse.cdt.make.core.enableAutoBuild</key>
					<value>false</value>
				</dictionary>
				<dictionary>
					<key>org.eclipse.cdt.make.core.enableCleanBuild</key>
					<value>true</value>
				</dictionary>
				<dictionary>
					<key>org.eclipse.cdt.make.core.enableFullBuild</key>
					<value>true</value>
				</dictionary>
				<dictionary>
					<key>org.eclipse.cdt.make.core.stopOnError</key>
					<value>true</value>
				</dictionary>
				<dictionary>
					<key>org.eclipse.cdt.make.core.useDefaultBuildCmd</key>
					<value>true</value>
				</dictionary>
			</arguments>
		</buildCommand>
		<buildCommand>
			<name>com.android.ide.eclipse.adt.ResourceManagerBuilder</name>
			<arguments>
			</arguments>
		</buildCommand>
		<buildCommand>
			<name>com.android.ide.eclipse.adt.PreCompilerBuilder</name>
			<arguments>
			</arguments>
		</buildCommand>
		<buildCommand>
			<name>org.eclipse.jdt.core.javabuilder</name>
			<arguments>
			</arguments>
		</buildCommand>
		<buildCommand>
			<name>com.android.ide.eclipse.adt.ApkBuilder</name>
			<arguments>
			</arguments>
		</buildCommand>
		<buildCommand>
			<name>org.eclipse.cdt.managedbuilder.core.ScannerConfigBuilder</name>
			<triggers>full,incremental,</triggers>
			<arguments>
			</arguments>
		</buildCommand>
	</buildSpec>
	<natures>
		<nature>com.android.ide.eclipse.adt.AndroidNature</nature>
		<nature>org.eclipse.jdt.core.javanature</nature>
		<nature>org.eclipse.cdt.core.cnature</nature>
		<nature>org.eclipse.cdt.core.ccnature</nature>
		<nature>org.eclipse.cdt.managedbuilder.core.managedBuildNature</nature>
		<nature>org.eclipse.cdt.managedbuilder.core.ScannerConfigNature</nature>
	</natures>
	<linkedResources>
		<link>
{java.sources}
		</link>
	</linkedResources>
</projectDescription>"""

	JavaSource = """<name>{name}</name>
			<type>2</type>
			<location>{path}</location>"""

	CProject = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<?fileVersion 4.0.0?>

<cproject storage_type_id="org.eclipse.cdt.core.XmlProjectDescriptionStorage">
	<storageModule moduleId="org.eclipse.cdt.core.settings">
		<cconfiguration id="com.android.toolchain.gcc.495145846">
			<storageModule buildSystemId="org.eclipse.cdt.managedbuilder.core.configurationDataProvider" id="com.android.toolchain.gcc.495145846" moduleId="org.eclipse.cdt.core.settings" name="Default">
				<externalSettings/>
				<extensions>
					<extension id="org.eclipse.cdt.core.ELF" point="org.eclipse.cdt.core.BinaryParser"/>
					<extension id="org.eclipse.cdt.core.VCErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.GmakeErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.CWDLocator" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.MakeErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.GCCErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.GASErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.GLDErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
				</extensions>
			</storageModule>
			<storageModule moduleId="cdtBuildSystem" version="4.0.0">
				<configuration artifactName="${ProjName}" buildProperties="" description="" id="com.android.toolchain.gcc.495145846" name="Default" parent="org.eclipse.cdt.build.core.emptycfg">
					<folderInfo id="com.android.toolchain.gcc.495145846.2099949476" name="/" resourcePath="">
						<toolChain id="com.android.toolchain.gcc.1075852585" name="com.android.toolchain.gcc" superClass="com.android.toolchain.gcc">
							<targetPlatform binaryParser="org.eclipse.cdt.core.ELF" id="com.android.targetPlatform.465101306" isAbstract="false" superClass="com.android.targetPlatform"/>
							<builder arguments="-j8" command="ndk-build" id="com.android.builder.773521748" keepEnvironmentInBuildfile="false" managedBuildOn="false" name="Android Builder" superClass="com.android.builder">
								<outputEntries>
									<entry flags="VALUE_WORKSPACE_PATH|RESOLVED" kind="outputPath" name="obj"/>
									<entry flags="VALUE_WORKSPACE_PATH|RESOLVED" kind="outputPath" name="libs"/>
								</outputEntries>
							</builder>
							<tool id="com.android.gcc.compiler.59181588" name="Android GCC Compiler" superClass="com.android.gcc.compiler">
								<inputType id="com.android.gcc.inputType.933235630" superClass="com.android.gcc.inputType"/>
							</tool>
						</toolChain>
					</folderInfo>
					<sourceEntries>
						<entry flags="VALUE_WORKSPACE_PATH|RESOLVED" kind="sourcePath" name="jni"/>
					</sourceEntries>
				</configuration>
			</storageModule>
			<storageModule moduleId="org.eclipse.cdt.core.externalSettings"/>
		</cconfiguration>
	</storageModule>
	<storageModule moduleId="cdtBuildSystem" version="4.0.0">
		<project id="{name}.null.346381687" name="{name}"/>
	</storageModule>
	<storageModule moduleId="scannerConfiguration">
		<autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>
		<scannerConfigBuildInfo instanceId="com.android.toolchain.gcc.495145846;com.android.toolchain.gcc.495145846.2099949476;com.android.gcc.compiler.59181588;com.android.gcc.inputType.933235630">
			<autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId="com.android.AndroidPerProjectProfile"/>
		</scannerConfigBuildInfo>
	</storageModule>
	<storageModule moduleId="refreshScope" versionNumber="1">
		<resource resourceType="PROJECT" workspacePath="/{name}"/>
	</storageModule>
	<storageModule moduleId="org.eclipse.cdt.make.core.buildtargets"/>
</cproject>"""


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os, shutil, plistlib, subprocess, glob

from ..Generator import Generator
from PBX         import Project
from ..Template  import Template

# class Xcode5
class Xcode5( Generator ):
	# ctor
	def __init__( self ):
		Generator.__init__( self )

		self.project  = None
		self.projects = {}

	# generate
	def generate( self ):
		Generator.generate( self )

		# Create project for each target
		self.forEachTarget( self.createProjects )

		# Setup projects
		for name, project in self.projects.items():
			target = project['target']
			pbx    = project['pbx']

			self.addTargetSources( target.name, target, pbx )
			self.addTargetLibraries( target.name, target, pbx )
			self.addTargetFrameworks( target.name, target, pbx )
			self.addTargetCommands( target.name, target, pbx )
			self.saveProject( project['project'], target )

		# Create workspace
		self.generateWorkspace()

	# listAvailableSDKs
	@staticmethod
	def listAvailableSDKs( platform ):
		# findXcode
		def findXcode():
			value = subprocess.check_output( ['xcode-select', '-p'] )

			# Found
			if value.find( '/Xcode.app/' ) != -1:
				return value.strip()

			# List applications
			apps = [path for path in glob.glob( '/Applications/Xcode*.app' ) if os.path.exists( os.path.join( path, 'Contents/MacOS/Xcode' ) )]
			if len( apps ) == 0:
				return None

			return os.path.join( apps[-1], 'Contents/Developer' )

		xcode  = findXcode()
		result = []
		path   = '{0}/Platforms/{1}.platform/Developer/SDKs/'.format( xcode, platform )

		# No SDK path found
		if not os.path.exists( path ):
			print 'Warning: no {0} SDK found'.format( platform )
			return [ platform ]

		# List SDKs
		for sdk in os.listdir( path ):
			with open( os.path.join( path, sdk, 'SDKSettings.plist' ), 'rb' ) as fp:
				result.append( plistlib.readPlist( fp )['CanonicalName'] )
				fp.close()

		return result

	# generateWorkspace
	def generateWorkspace( self ):
		# Create the workspace folder
		workspace = os.path.join( self.binaryDir, self.sourceProject.name + '.xcworkspace' )

		# Create folder
		if not os.path.exists( workspace ):
			os.mkdir( workspace )

		# Generate project list
		projects = ''

		for name, project in self.projects.items():
			folder    = self.getPathForTarget( project['target'] )
			projects += "<FileRef location='group:{0}'/>\n".format( os.path.join( folder, name + '.xcodeproj') )

		# Dump workspace to file
		Template( Xcode5.Workspace ).compileToFile( os.path.join( workspace, 'contents.xcworkspacedata' ), { 'projects': projects } )

	# saveProject
	def saveProject( self, project, target ):
		# Create project folder
		folder = os.path.join( self.getPathForTarget( target ), target.name + '.xcodeproj' )

		if not os.path.exists( folder ):
			os.mkdir( folder )

		# Dump project to disk
		file = open( os.path.join( folder, 'project.pbxproj' ), 'wt' )
		file.write( project.generate() )
		file.close()

	# createProjects
	def createProjects( self, name, target ):
		settings = {
			'Debug':    self.generateConfiguration( target, 'Debug' ),
		    'Release':  self.generateConfiguration( target, 'Release' )
		}

		# Create project
		project = Project( target.name, self.getProjectSettings() )
		pbx     = None

		# Create target
		if target.type == 'executable':
			if target.params and target.params['bundle']:
				pbx = project.addApplicationBundle( name, name + '.app', settings, 'Debug', self.makefile.get( 'DEVELOPMENT_TEAM' ) )
				self.addInfoPlist( target, pbx )
				if target.resources:
					self.addResources( target, pbx )
			else:
				pbx = project.addApplication( name, name, settings, 'Debug' )
		else:
			pbx = project.addStaticLibrary( name, 'lib' + name + '.a', settings, 'Debug' )

		self.projects[name] = { 'project': project, 'target': target, 'pbx': pbx }

	# getProjectSettings
	def getProjectSettings( self ):
		return None

	# getPlatformId
	def getPlatformId( self ):
		return None

	# addInfoPlist
	def addInfoPlist( self, target, pbx ):
		pbx.addPlist( 'Info.plist' )
		Template( Xcode5.Plist ).compileToFile( os.path.join( self.getPathForTarget( target ), 'Info.plist' ), {
																										'identifier':       self.makefile.get( 'IDENTIFIER' ),
																										'version':          self.makefile.get( 'VERSION' ),
																										'short.version':    self.makefile.get( 'SHORT_VERSION' ),
		                                                                                                'facebook.app.id':  self.makefile.get( 'FACEBOOK_APP_ID' )
																									} )

	# addResources
	def addResources( self, target, pbx ):
		icons  = os.path.join( target.sourcePath, target.resources, 'images', 'icons.'  + self.getPlatformId() )
		launch = os.path.join( target.sourcePath, target.resources, 'images', 'launch.' + self.getPlatformId() )
		assets = os.path.join( target.sourcePath, target.resources, 'assets', 'assets.dpk' )

		# Icons
		if os.path.exists( icons ):
			dst = os.path.join( self.getPathForTarget( target ), 'icons.xcassets' )
			if os.path.exists( dst ):
				shutil.rmtree( dst )

			shutil.copytree( icons, dst )

			pbx.addAssetCatalog( 'icons.xcassets' )
			pbx.setAppIcon( 'Icon' )

		# Launch images
		if os.path.exists( launch ):
			dst = os.path.join( self.getPathForTarget( target ), 'launch.xcassets' )
			if os.path.exists( dst ):
				shutil.rmtree( dst )

			shutil.copytree( launch, dst )

			pbx.addAssetCatalog( 'launch.xcassets' )
			pbx.setLaunchImage( 'Launch' )

		# Assets
		if os.path.exists( assets ):
			dst = os.path.join( self.getPathForTarget( target ), 'assets.dpk' )

			shutil.copyfile( assets, dst )

			pbx.addResourceFile( 'assets.dpk' )

	# addTargetCommands
	def addTargetCommands( self, name, target, pbx ):
		# No commands for target
		if len( target.commands ) == 0:
			return

		# Add shell script
		pbx.addShellScript( 'make -C {0} -f {1}.commands'.format( self.getPathForTarget( target ), name ) )

		# Generate makefile
		commands = self.processEachTargetCommand( target, self.compileCommand )
		depends  = self.processEachTargetCommand( target, self.compileDependency )
		Template( Xcode5.CodegenScript ).compileToFile( os.path.join( self.getPathForTarget( target ), name + '.commands' ), { 'depends': depends, 'commands': commands } )

	# addTargetSources
	def addTargetSources( self, name, target, pbx ):

		# addSourceFile
		def addSourceFile( file ):
			addSourceFile.target.addSourceFile( file.fullPath, file.sourceFolder )

		# Add source files to target
		addSourceFile.target = pbx
		self.forEachTargetSource( target, addSourceFile )

	# addTargetFrameworks
	def addTargetFrameworks( self, name, target, pbx ):

		# addFramework
		def addFramework( target, name, library ):
			if name.endswith( '.framework' ) and not os.path.isabs( name ):
				name = os.path.join( self.sourceDir, name )

			addFramework.target.addFramework( name )

		# Add global frameworks
		addFramework.target = pbx
		self.forEachTargetFramework( target, addFramework )
		
	# addTargetLibraries
	def addTargetLibraries( self, name, target, pbx ):

		# addLibrary
		def addLibrary( library ):
			if library.type == 'local':
				if library.name in self.projects.keys():
					addLibrary.target.addProjectLibrary( self.projects[library.name]['pbx'] )
				else:
					addLibrary.target.addLibrary( 'lib' + library.name + '.a' )
			elif library.type == 'package':
				addLibrary.target.addLibrary( os.path.join( library.libs, library.fileName ) )
				for item in library.items:
					addLibrary.target.addLibrary( item )
			else:
				addLibrary.target.addLibrary( os.path.join( library.libs, library.fileName ) )

		# Add linked libraries
		addLibrary.target = pbx
		self.forEachTargetLibrary( target, addLibrary )

	# compileCommand
	def compileCommand( self, target, cmd ):
		output = ''
		for ext in cmd.generatedExtensions:
			output += cmd.output + ext + ' '

		input = ''
		for fileName in cmd.input:
			input += fileName + ' '

		return Template( Xcode5.CodegenCommand ).compile( { 'output': output, 'input': input, 'command': cmd.command, 'message': cmd.message } )

	# compileDependency
	def compileDependency( self, target, cmd ):
		result = ''

		for ext in cmd.generatedExtensions:
			result += '\\\n\t' + cmd.output + ext

		return result

	# generateConfiguration
	def generateConfiguration( self, target, name ):
		# generateHeaderIncludePaths
		def generateHeaderIncludePaths( target, path ):
			return ' ' + path + ' '

		# generateLibrarySearchPaths
		def generateLibrarySearchPaths( library ):
			return ' ' + library.libs + ' ' if library.type == 'external' else None

		# generateDefines
		def generateDefines( target, define ):
			return ' ' + define + ' '

		paths   = self.processEachTargetInclude( target, generateHeaderIncludePaths ).strip().split( ' ' )
		defines = self.processEachTargetDefine( target, generateDefines ).strip().split( ' ' )
		libs    = self.processEachTargetLib( target, generateLibrarySearchPaths ).strip().split( ' ' )
		paths   = set( paths )
		paths   = list( paths )
		libs    = set( libs )
		libs    = list( libs )
		libs.append( '$(inherited)' )
		paths.append( '$(inherited)' )

		return { 'PRODUCT_NAME': '"$(TARGET_NAME)"', 'HEADER_SEARCH_PATHS': paths, 'LIBRARY_SEARCH_PATHS': libs, 'GCC_PREPROCESSOR_DEFINITIONS': defines }

	CodegenCommand = """
{output}: {input}
	@echo '{message}'
	@{command}
	"""

	CodegenScript = """
all: {depends}

{commands}
"""

	Plist = """
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
	<key>CFBundleIdentifier</key>
	<string>{identifier}</string>
	<key>CFBundleShortVersionString</key>
	<string>{short.version}</string>
	<key>CFBundleVersion</key>
	<string>{version}</string>
	<key>CFBundleVersion</key>
	<string>{version}</string>
	<key>CFBundleExecutable</key>
	<string>${PRODUCT_NAME}</string>
	<key>FacebookAppID</key>
	<string>{facebook.app.id}</string>
</dict>
</plist>
"""

	Workspace = """<?xml version='1.0' encoding='UTF-8'?>
<Workspace version='1.0'>
{projects}
</Workspace>
"""


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os

# class Template
class Template:
	# ctor
	def __init__( self, template ):
		self.template 	= template.encode( 'utf-8' )
		self.result		= ''
		self.isCompiled	= False

	# compile
	def compile( self, args, convertLineEndings = False ):
		self.result 	= self.template
		self.isCompiled = True

		if args == None:
			return self.result
		
		for key, value in args.items():
			self.result = self.result.replace( '{' + key + '}', str( value ) )
		
		if convertLineEndings:
			self.result = self.result.replace( '\r', '' )
			self.result = self.result.replace( '\n', os.linesep )
				
		return self.result

	# compileToFile
	def compileToFile( self, fileName, args = None ):
		compiled = self.compile( args )

		file = open( fileName, 'w' )
		file.write( compiled )
		file.close()

	# output
	def output( self, file, args = None ):
		if not self.isCompiled:
			self.compile( args )

		file.write( self.result )


#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from FindLibrary import findLibrary

# class Library
class Library:
	# ctor
	def __init__( self, target, name ):
		self._target    = target
		self._name      = name

	# type
	@property
	def type( self ):
		return 'unknown'

	# __str__
	def __str__( self ):
		return "[Library {0}]".format( self._name )

	# name
	@property
	def name( self ):
		return self._name


# class LocalLibrary
class LocalLibrary( Library ):
	# ctor
	def __init__( self, target, name ):
		Library.__init__( self, target, name )

	# type
	@property
	def type( self ):
		return 'local'

	# __str__
	def __str__( self ):
		return "[LocalLibrary {0}]".format( self._name )


# class ExternalLibrary
class ExternalLibrary( Library ):
	# ctor
	def __init__( self, name, headers, libs ):
		Library.__init__( self, None, name )

		self._headers   = headers
		self._libs      = libs

	# headers
	@property
	def headers( self ):
		return self._headers

	# libs
	@property
	def libs( self ):
		return self._libs

	# fileName
	@property
	def fileName( self ):
		return self.getLibraryFileName( self._name )

	# type
	@property
	def type( self ):
		return 'external'

	# __str__
	def __str__( self ):
		return "[{0} headers='{1}' libs={2}]".format( self._name, self._headers, self._libs )

	# getLibraryFileName
	def getLibraryFileName( self, name ):
		staticLibraryPath = os.path.join( self._libs, 'lib' + name + '.a' )
		sharedLibraryPath = os.path.join( self._libs, 'lib' + name + '.dylib' )

		if os.path.exists( sharedLibraryPath ): return sharedLibraryPath

		return staticLibraryPath

	# find
	@staticmethod
	def find( name ):
		info = findLibrary( name )
		if info:
			return ExternalLibrary( info['name'], info['headers'], info['library'] )

		return None

# class ExternalPackage
class ExternalPackage( ExternalLibrary ):
	# ctor
	def __init__( self, name, headers, libs, items ):
		ExternalLibrary.__init__( self, name, headers, libs )

		self._items = items

	# type
	@property
	def type( self ):
		return 'package'

	# items
	@property
	def items( self ):
		return [self.getLibraryFileName( item ) for item in self._items]

	# __str__
	def __str__( self ):
		return "[{0} headers='{1}' libs='{2}' items='{3}']".format( self._name, self._headers, self._libs, self._items )

	# find
	@staticmethod
	def find( name, *items ):
		info = findLibrary( name )
		if info:
			return ExternalPackage( info['name'], info['headers'], info['library'], items )

		return None

# class Framework
class Framework( Library ):
	# ctor
	def __init__( self, target, name ):
		Library.__init__( self, target, name )

	# __str__
	def __str__( self ):
		return "[Framework {0}]".format( self._name )

	# type
	@property
	def type( self ):
		return 'framework'


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os, string

from Make       import Make
from ..Template import Template

# class Android
class Android( Make ):
	# constructor
	def __init__( self ):
		Make.__init__( self )
		self.targetType = { 'static': 'BUILD_STATIC_LIBRARY', 'shared': 'BUILD_SHARED_LIBRARY', 'executable': 'BUILD_SHARED_LIBRARY'  }

	# getPathForTarget
	def getPathForTarget( self, target ):
		return self.binaryDir + '/jni/' + target.name + '.dir'

	# generateIncludePath
	def generateIncludePath( self, target, path ):
		return string.replace( os.path.relpath( path, self.binaryDir ), '\\', '/' )
		
	# generate
	def generate( self ):
		print( 'Generating Android NDK project...' )
	
		# JNI folder
		if not os.path.exists( self.binaryDir + '/jni' ):
			os.makedirs( self.binaryDir + '/jni' )

		# Generate targets
		self.forEachTarget( self.generateTarget )

		# Generate root makefile
		self.generateProjectMakefile()

	# generateTarget
	def generateTarget( self, name, target ):
		# generateLibDependency
		def generateLibDependecy( target, name, lib ):
			return name + ' '

		#Create folder
		path 		= '{0}/jni/{1}.dir'.format( self.binaryDir, target.name )
		fileName 	= os.path.join( path, 'Android.mk' )

		if not os.path.exists( path ):
			os.makedirs( path )

		libs 			= self.processEachTargetLib( target, generateLibDependecy )
		sources 		= 'Autogenerated.c ' + self.processEachTargetSource( target, ['.c', '.cpp'], self.generateFileName )
		flags 			= self.getCompilerFlagsForTarget( self.sourceProject ) + ' ' + self.getCompilerFlagsForTarget( target )
		shared  		= self.processEachTargetFramework( target, self.generateFramework )
		commandRules 	= self.processEachTargetCommand( target, self.generateCommandRule )
		commands 		= self.processEachTargetCommand( target, self.generateCommand )

		# Write Android.mk
		Template( Android.Target ).compileToFile( fileName, {
																'name': 			target.name,
																'touch':			'echo // Autogenerated  >' if os.name == 'nt' else 'touch',
																'libs': 			libs,
																'sources': 			sources,
																'type': 			self.targetType[target.type],
																'flags': 			flags,
																'shared': 			shared,
																'command.rules':	commandRules,
																'commands':			commands,
																'command.proxy':	os.path.join( self.getPathForTarget( target ), 'Autogenerated.c' ),
															} )

	# generateFramework
	def generateFramework( self, target, name, lib ):
		return '-l' + name + ' '

	# generateCommand
	def generateCommand( self, target, cmd ):
		output = ''
		for ext in cmd.generatedExtensions:
			output += '$(LOCAL_PATH)/' + self.convertPath( cmd.output, self.getPathForTarget( target ) ) + ext + ' '

		return output

	# generateCommandRule
	def generateCommandRule( self, target, cmd ):
		targetPath = self.getPathForTarget( target )
		baseName   = os.path.basename( cmd.input[0] )

		input = ''
		for fileName in cmd.input:
			input += self.convertPath( fileName, self.binaryDir ) + ' '

		output = ''
		for ext in cmd.generatedExtensions:
			output += '$(LOCAL_PATH)/' + self.convertPath( cmd.output, targetPath ) + ext + ' '

		return Template( Android.Command ).compile( {
														'message': 	cmd.message,
														'input': 	input,
														'output': 	output,
														'command': 	cmd.command,
														'target': 	target.name,
														'baseName': baseName,
														'proxy':	os.path.join( targetPath, output ),
													} )

	# generateFileName
	def generateFileName( self, target, filePath, baseName, ext ):
		return filePath + ' '

	# generateProjectMakefile
	def generateProjectMakefile( self ):
		# addTarget
		def addTarget( name, target ):
			return 'include $(TOP_PATH)/{0}.dir/Android.mk\n'.format( name )

		# Generate targets
		targets = self.processEachTarget( addTarget )

		# Write root Android.mk & Application.mk
		Template( Android.Makefile ).compileToFile( '{0}/jni/Android.mk'.format( self.binaryDir ), { 'targets': targets } )
		Template( Android.Application ).compileToFile( '{0}/jni/Application.mk'.format( self.binaryDir ) )

	# Project makefile
	Makefile = """
# Autogenerated by YAP

TOP_PATH:=$(call my-dir)
{targets}

$(call import-module,android/native_app_glue)
	"""

	# Application.mk
	Application = """
# Autogenerated by YAP

APP_STL 		:= gnustl_static
APP_PLATFORM 	:= android-9
APP_ABI 		:= armeabi
APP_CFLAGS		:= -O3 -w -fno-exceptions -fno-rtti
	"""

	# Target makefile
	Target = """
# Autogenerated by YAP

LOCAL_PATH:= $(call my-dir)

include $(CLEAR_VARS)

LOCAL_SHORT_COMMANDS	:= false
LOCAL_MODULE    		:= {name}
LOCAL_SRC_FILES 		:= {sources}
LOCAL_CFLAGS			+= {flags}
LOCAL_STATIC_LIBRARIES	:= android_native_app_glue {libs}
LOCAL_ARM_MODE 			:= arm
LOCAL_LDLIBS			:= {shared}

$(LOCAL_PATH)/Autogenerated.c: {commands}
	@{touch} {command.proxy}

{command.rules}

include $({type})
"""

	Command = """
{output}: {input}
	$(info Shell    : {target} <= {message})
	@{command}
"""


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os
import string

class Generator:
	# ctor
	def __init__( self ):
		self.sourceProject = None
		self.binaryDir     = None
		self.sourceDir     = None
		self.makefile      = None

	# getPathForTarget
	def getPathForTarget( self, target ):
		return os.path.join( self.binaryDir, target.name + '.dir' )

	# initialize
	def initialize( self, makefile, source, binary, project ):
		self.makefile       = makefile
		self.sourceDir      = source
		self.binaryDir      = binary
		self.sourceProject  = project

	# generate
	def generate( self ):
		# Create project folder
		if not os.path.exists( self.binaryDir ):
			os.makedirs( self.binaryDir )

		self.forEachTarget( self.generateTarget )

	# generateTarget
	def generateTarget( self, name, target ):
		path = self.getPathForTarget( target )

		# Create target folder
		if not os.path.exists( path ):
			os.makedirs( path )

		if not os.path.exists( path + '/obj' ):
			os.makedirs( path + '/obj' )

	# processEachTarget
	def processEachTarget( self, processor, filter = None ):
		# callback
		def callback( name, target ):
			callback.result += processor( name, target )

		callback.result = ''
		self.forEachTarget( callback, filter )

		return callback.result

	# processEachTargetSource
	def processEachTargetSource( self, target, filter, processor ):
		# callback
		def callback( filePath, baseName, ext ):
			callback.result += processor( target, filePath, baseName, ext )

		callback.result = ''
		self.forEachTargetSource( target, callback, filter )

		return callback.result

	# processEachGroup
	def processEachGroup( self, target, processor ):
		# callback
		def callback( path, files ):
			callback.result += processor( target, path, files )

		callback.result = ''
		self.forEachGroup( target, callback )

		return callback.result

	# processEachTargetInclude
	def processEachTargetInclude( self, target, processor ):
		# callback
		def callback( path ):
			callback.result += processor( target, path )

		callback.result = ''
		self.forEachTargetInclude( target, callback )

		return callback.result

	# processEachTargetDefine
	def processEachTargetDefine( self, target, processor ):
		# callback
		def callback( define ):
			callback.result += processor( target, define )

		callback.result = ''
		self.forEachTargetDefine( target, callback )

		return callback.result

	# processEachTargetCommand
	def processEachTargetCommand( self, target, processor ):
		# callback
		def callback( target, command ):
			result = processor( target, command )

			if result != None:
				callback.result += result

		callback.result = ''

		for command in target.commands:
			callback( target, command )

		return callback.result

	# processEachTargetLib
	def processEachTargetLib( self, target, processor ):
		# callback
		def callback( library ):
			result = processor( library )

			if result != None:
				callback.result += result

		callback.result = ''
		self.forEachTargetLibrary( target, callback )

		return callback.result

	# processEachTargetFramework
	def processEachTargetFramework( self, target, processor ):
		# callback
		def callback( target, name, lib ):
			result = processor( target, name, lib )

			if result != None:
				callback.result += result

		callback.result = ''
		self.forEachTargetFramework( target, callback )

		return callback.result

	# forEachTargetSource
	def forEachTargetSource( self, target, callback, filter = None ):
		for file in target.filterSourceFiles( filter ):
			callback( file )
	#	path = self.getPathForTarget( target )

	#	for source in target.sources:
		#	fileName		= os.path.basename( source )
		#	baseName, ext 	= os.path.splitext( fileName )
		#	filePath		= string.replace( os.path.relpath( source, path ), '\\', '/' )
	#		name, ext = source.nameAndExtension

	#		if filter == None or ext in filter:
	#			callback( source.relativePath( path ), name, ext )

	# forEachTarget
	def forEachTarget( self, callback, filter = None ):
		for target in self.sourceProject.filterTargets( filter ):
			callback( target.name, target )
	#	for name, target in self.sourceProject.targets.items():
	#		if filter == None or filter == target.type:
	#			callback( name, target )

	# forEachTargetLibrary
	def forEachTargetLibrary( self, target, callback ):
		for lib in target.filterLibraries():
			callback( lib )
		# Get libs
	#	libs = []
	#	for lib in self.getLibsForTarget( target ):
	#		if not lib in libs:
	#			libs.append( lib )

	#	for name in libs:
	#		callback( target, name, self.sourceProject.findTarget( name, [Target.StaticLibrary, Target.SharedLibrary] ) )

	# forEachTargetFramework
	def forEachTargetFramework( self, target, callback ):
		for framework in target.filterFrameworks():
			callback( target, framework.name, None )
		#	if lib['framework']:
		#		callback( target, lib['name'], None )

	# findTargetByName
	def findTargetByName( self, name ):
		if name in self.sourceProject.targets.keys():
			return self.sourceProject.targets[name]

		return None

	# getLibsForTarget
	def getLibsForTarget( self, target ):
		result = []

		for library in target.libraries:
			result.append( library )
		#	name = lib['name']

		#	if not lib['framework']:
		#		result.append( name )

		#	if not name in self.sourceProject.targets.keys():
		#		continue

		#	lib = self.sourceProject.targets[name]

		#	if lib.type == 'static':
		#		libs = self.getLibsForTarget( lib )
		#		for name in libs:
		#			result.append( name )

		return result

	# forEachTargetInclude
	def forEachTargetInclude( self, target, callback ):
		path = self.getPathForTarget( target )

		# Project include paths
		if target != self.sourceProject:
			self.forEachTargetInclude( self.sourceProject, callback )

		# Target
		for include in target.includes:
			if os.path.isabs( path ):
				include = string.replace( include, '\\', '/' )
			else:
				include = string.replace( os.path.relpath( include, path ), '\\', '/' )

			callback( include )

		# Libraries
		for library in target.filterLibraries( lambda lib: lib.type == 'external' ):
			callback( library.headers )

	# forEachTargetDefine
	def forEachTargetDefine( self, target, callback ):
		# Project defines
		if target != self.sourceProject:
			self.forEachTargetDefine( self.sourceProject, callback )

		# Target
		for define in target.defines:
			callback( define )

	# forEachCommand
	def forEachCommand( self, target, callback ):
		for cmd in target.commands:
			callback( cmd )

	# forEachGroup
	def forEachGroup( self, target, callback ):
		# iterator
		def iterator( group ):
			for name in group.groups:
				iterator( group.groups[name] )

			path = os.path.relpath( group.path, target.name )
			if path == '.':
				return

			callback( path, group.files )

		iterator( target.groups )



#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

from ObjectList     import ObjectList
from Resource       import Resource
from ...Template    import Template

# class Group
class Group( Resource ):
	# ctor
	def __init__( self, objects, name ):
		Resource.__init__( self, 'PBXGroup', name )

		self.objects  = objects
		self.children = ObjectList( 'PBXGroup' )

	# add
	def add( self, item ):
		self.children.add( item )

	# resolveGroup
	def resolveGroup( self, name ):
		for group in self.children.items:
			if group.name == name:
				return group

		group = self.objects.createGroup( name, self )
		return group

	# compile
	def compile( self ):
		name = 'name = {0};'.format( self.name ) if self.name else ''
		return Template( Group.Root ).compile( { 'id': self.id, 'isa': self.isa, 'name': self.name, 'name.property': name, 'children': self.children.compileList() } )

	Nested = "\t\t\t\t{id} /* {name} */,\n"

	Root = """
		{id} /* {name} */ = {
			isa = {isa};
			children = (
{children}
			);
			{name.property}
			sourceTree = "<group>";
		};
"""


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import string, os

from ..Generator import Generator
from ..Template  import Template

# class Make
class Make( Generator ):
	# ctor
	def __init__( self ):
		Generator.__init__( self )

		self.globalFlags = ''
		self.extensions  = { 'shared': 'so', 'static': 'a', 'executable': '' }
		self.toolchain   = { 'cc': 'gcc', 'cxx': 'g++', 'ar': 'ar' }

	# generateTargetDependencies
	def generateTargetDependencies( self, target ):
		result = ''
		
		for name in target.libs:
			result += name + ' '
			
		return result

	# generateIncludePath
	def generateIncludePath( self, target, path ):
		basePath = self.getPathForTarget( target )
		return string.replace( os.path.relpath( path, basePath ), '\\', '/' )

	# generate
	def generate( self ):
		self.globalFlags = self.getCompilerFlagsForTarget( self ) + ' -w '
		
		Project.generate( self )
		self.generateProjectMakefile()

	# generateProjectMakefile
	def generateProjectMakefile( self ):
		clean 			= ''
		compileTargets 	= ''
		targets 		= ''

		# Clean targets
		for name in self.targets:
			target 	=  self.targets[name]
			clean 	+= Template( Make.CleanTarget ).compile( { 'name': name } )
			targets += name + ' '

		# Compile targets
		for name in self.targets:
			target = self.targets[name]
			compileTargets += Template( Make.CompileTarget ).compile( { 'name': name, 'dependencies': self.generateTargetDependencies( target ) } )

		file = open( Makefile.BinaryDir + '/Generator', 'w' )
		file.write( Template( Make.Makefile ).compile( { 'clean': clean, 'compileTargets': compileTargets, 'targets': targets } ) )
		file.close()

	# generateTarget
	def generateTarget( self, target ):
		Project.generateTarget( self, target )
		
		makefile = '{0}/{1}.dir/Generator'.format( Makefile.BinaryDir, target.name )

		product   	= self.getProductForTarget( target )
		linkflags 	= self.getLinkerFlagsForTarget( target )
		cxxflags  	= self.getFlagsForTarget( target )
		linker 	  	= self.getLinkerForTarget( target )

		cxxo 	  	= self.processEachTargetSource( target, ['.c', '.cpp'], self.generateObjectFileList )
		rules 	  	= self.processEachTargetSource( target, ['.c', '.cpp'], self.generateCompilerRule )

		codegenout	= self.processEachTargetCommand( target, self.generateCodegenOutput )
		gencxxo		= self.processEachTargetCommand( target, self.generateCodegenObject )
		commands  	= self.processEachTargetCommand( target, self.generateCommandRule )

		Template( Make.Target ).compileToFile( makefile, { 'product': product, 'target': target.name, 'linker.flags': linkflags, 'cxx.flags': cxxflags,
																	  'gencxx.o': gencxxo, 'commands': commands, 'codegen.output': codegenout,
																	  'cxx.o': cxxo, 'rules': rules, 'linker': linker,
																	  'tool.cc': self.toolchain['cc'], 'tool.cxx': self.toolchain['cxx'], 'tool.ar': self.toolchain['ar'] } )

	# generateObjectFileList
	def generateObjectFileList( self, target, filePath, baseName, ext ):
		return '$(OBJ_DIR)/' + baseName + '.o '

	# generateCompilerRule
	def generateCompilerRule( self, target, filePath, baseName, ext ):
		return Template( Make.CxxRule ).compile( { 'compiler': '$(CXX)' if ext == '.cpp' else '$(CC)', 'filePath': filePath, 'baseName': baseName, 'target': target.name } )

	# generateAutogeneratedCompilerRule
	def generateAutogeneratedCompilerRule( self, target, cmd ):
		targetPath = self.getPathForTarget( target )
		name = self.convertPath( cmd.output, targetPath )
		return self.generateCompilerRule( target, name + '.cpp', name, '.cpp' )

	# generateCommandRule
	def generateCommandRule( self, target, cmd ):
		targetPath = self.getPathForTarget( target )
		baseName   = os.path.basename( cmd.input[0] )

		input = ''
		for fileName in cmd.input:
			input += self.convertPath( fileName, targetPath ) + ' '

		output = ''
		for ext in cmd.generatedExtensions:
			output += self.convertPath( cmd.output, targetPath ) + ext + ' '

		return Template( Make.CxxGenerate ).compile( {
																	'message': 	cmd.message,
																	'input': 	input,
																	'output': 	output,
																	'command': 	cmd.command.replace( '\\', '/' ),
																	'target': 	target.name,
																	'baseName': baseName
																} )

	# generateCodegenOutput
	def generateCodegenOutput( self, target, cmd ):
		result = ''

		for ext in cmd.generatedExtensions:
			result += self.convertPath( cmd.output, self.getPathForTarget( target ) ) + ext + ' '

		return result

	# generateCodegenObject
	def generateCodegenObject( self, target, cmd ):
		return '$(OBJ_DIR)/' + os.path.basename( cmd.output ) + '.o '

	# getProductForTarget
	def getProductForTarget( self, target, forLinker = False ):
		ext  = self.getExtensionForTarget( target )
		name = target.name

		if len( ext ) == 0:
			return name

		if target.type == 'static':
			name = 'lib' + name

		return name + '.' + ext

	# getExtensionForTarget
	def getExtensionForTarget( self, target ):
		if target.type in self.extensions:
			return self.extensions[target.type]

		print "Make::getExtensionForTarget : unknown target type '{0}'".format( target.type )
		return ''

	# getLinkerForTarget
	def getLinkerForTarget( self, target ):
		if target.type == 'shared':
			return '$(CXX) $(CXX_OBJ) -o $(PRODUCT) $(LINK_FLAGS)'
		elif target.type == 'executable':
			return '$(CXX) $(CXX_OBJ) -o $(PRODUCT) $(LINK_FLAGS)'
		else:
			return '$(AR) rs $(PRODUCT) $(LINK_FLAGS) $(CXX_OBJ)'

		print "Make::getLinkerForTarget : unknown target type '{0}'".format( target.type )
		return ''

	# getLibrariesForTarget
	def getLibrariesForTarget( self, target ):
		result 		= ''
		currentPath	= self.getPathForTarget( target )
		
		for name in target.libs:
			lib  = self.targets[name]
			path = self.getPathForTarget( lib )
			name = self.getProductForTarget( lib, True )

			result += self.convertPath( path + '/lib' + name, currentPath ) + ' '
			result += self.getLibrariesForTarget( lib )
			
		return result

	# getLinkerFlagsForTarget
	def getLinkerFlagsForTarget( self, target ):
		if target.type != 'static':
			return self.getLibrariesForTarget( target )

		return ''

	# getCompilerFlagsForTarget
	def getCompilerFlagsForTarget( self, target ):
		result = ' '
		
		# Defines
		for define in target.defines:
			result += '-D' + define + ' '

		# Include paths
		for include in target.includes:
			path    = self.sourcePath if include == '.' else include
			result += '-I' + self.generateIncludePath( target, path ) + ' '
			
		return result

	# getFlagsForTarget
	def getFlagsForTarget( self, target ):
		return self.globalFlags + self.getCompilerFlagsForTarget( target )


	# convertPath
	def convertPath( self, path, base = None ):
		if base == None:
			base = os.getcwd()

		return string.replace( os.path.relpath( path, base ), '\\', '/' )

	############################### TEMPLATES

	Makefile = """
# Autogenerated by YAP

all: folders {targets}
	@echo "Done"

folders:
	mkdir -p Debug Release

{compileTargets}

clean:
{clean}
"""

	Target = """
# Autogenerated by YAP

CC  = {tool.cc}
CXX = {tool.cxx}
AR  = {tool.ar}

############################## TARGET

PRODUCT		= {product}
OBJ_DIR 	= obj

LINK_FLAGS 	= {linker.flags}
CXX_FLAGS 	= {cxx.flags}
CXX_OBJ 	= {cxx.o}

all: folders commands $(PRODUCT)
	@true

folders:
	@mkdir -p $(OBJ_DIR)
		
clean:
	@echo "Cleaning {target}..."
	@rm -f obj/*.o *.a *.bc *.html *.h *.cpp

$(PRODUCT): $(CXX_OBJ)
	@echo 'Linking...'
	@{linker}
	@cp $(PRODUCT) ../Release/$(PRODUCT)

############################## COMMANDS RULES

commands: {codegen.output}
	@true

{commands}

############################## CXX RULES

{rules}
"""
	CxxRule = """
$(OBJ_DIR)/{baseName}.o: {filePath}
	@echo '{target} <= {baseName}'
	@{compiler} $(CXX_FLAGS) -c $< -o $@
"""

	CxxGenerate = """
{output}: {input}
	@echo '{target} <= {baseName} {message}'
	@{command}
"""

	CleanTarget = "\t@cd {name}.dir && make clean\n"

	CompileTarget = """
{name}: {dependencies}
	@echo "Compiling {name}..."
	@cd {name}.dir && make -f Generator
	"""
